"""Copyright © 2023 Pixelgen Technologies AB."""

import logging
import time
from functools import partial
from typing import Iterable, get_args

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from scipy.stats import norm
from statsmodels.stats.multitest import multipletests

from pixelator.common.statistics import correct_pvalues, wilcoxon_test
from pixelator.common.utils import get_pool_executor
from pixelator.mpx.analysis.analysis_engine import PerComponentAnalysis
from pixelator.mpx.analysis.permute import permutations
from pixelator.mpx.analysis.polarization.types import PolarizationTransformationTypes
from pixelator.mpx.graph.utils import Graph
from pixelator.mpx.pixeldataset import MIN_VERTICES_REQUIRED, PixelDataset

logger = logging.getLogger(__name__)


def _compute_for_marker(
    marker, marker_count_matrix, weight_matrix, normalization_factor
):
    try:
        x_centered = np.array(
            marker_count_matrix[marker] - marker_count_matrix[marker].mean()
        )
        x_centered_squared_sum = (x_centered * x_centered).sum()
        element_wise_weighted_sum = (x_centered @ weight_matrix @ x_centered).sum()
        r = normalization_factor * (element_wise_weighted_sum / x_centered_squared_sum)
        return r
    except ValueError as e:
        logger.warning("Error computing Moran's I for marker %s: %s", marker, e)
        return np.nan


def _compute_morans_i(
    marker: str,
    marker_count_matrix: pd.DataFrame,
    marker_count_matrix_permuted: Iterable[pd.DataFrame],
    weight_matrix: csr_matrix,
    normalization_factor: float,
):
    r = _compute_for_marker(
        marker, marker_count_matrix, weight_matrix, normalization_factor
    )
    perm_rs = np.fromiter(
        (
            _compute_for_marker(marker, perm, weight_matrix, normalization_factor)
            for perm in marker_count_matrix_permuted
        ),
        dtype=np.float64,
    )
    perm_rs_mean = np.nanmean(perm_rs)
    perm_rs_std_dev = np.nanstd(perm_rs)
    perm_rs_z_score = (r - perm_rs_mean) / perm_rs_std_dev
    p_value = norm.sf(np.abs(perm_rs_z_score))
    return marker, r, perm_rs_z_score, perm_rs_mean, perm_rs_std_dev, p_value


def polarization_scores_component_graph(
    graph: Graph,
    component_id: str,
    transformation: PolarizationTransformationTypes = "log1p",
    n_permutations: int = 50,
    min_marker_count: int = 2,
    random_seed: int | None = None,
) -> pd.DataFrame:
    """Calculate Moran's I statistics for a component graph.

    Computes polarization statistics for all antibodies in the `graph` given
    as input (a single connected component). The statistics are computed using
    Moran's I autocorrelation to measure how clustered/localized the spatial
    patterns of the antibody is in the graph.

    The statistics contain the I value, as well as a p-value, the adjusted p-value
    and the z-score.The later are based on permuting the input graph to estimate a
    null-hypothesis for the Moran's I statistic.

    The function returns a pd.DataFrame with the following columns:
      morans_i, morans_p_value, morans_z, marker, component

    :param graph: a graph (it must be a single connected component)
    :param component_id: the id of the component
    :param transformation: the count transformation method to use (raw, log1p)
    :param n_permutations: the number of permutations to use to estimate the
                           null-hypothesis for the Moran's I statistic
    :param min_marker_count: the minimum number of counts of a marker to calculate
                             the Moran's I statistic
    :param random_seed: the random seed to use to ensure that the permutations
                        are reproducible across runs
    :returns: a pd.DataFrame with the polarization statistics for each antibody
    :rtype: pd.DataFrame
    :raises: AssertionError when the input is not valid
    """
    if graph.vcount() < MIN_VERTICES_REQUIRED:
        logger.debug(
            (
                "Trying to compute polarization scores in component %s that has less"
                " than %i vertices which is the minimum required"
            ),
            component_id,
            MIN_VERTICES_REQUIRED,
        )
        return pd.DataFrame()

    def _compute_morans_on_current_graph():
        N = float(graph.vcount())

        W = graph.get_adjacency_sparse()
        # Normalize the weights by the degree of the node
        W = W / W.sum(axis=0)
        X = graph.node_marker_counts

        # Calculate normalization factor
        C = N / W.sum()

        # Record markers to keep
        # Remove markers with zero variance and markers below minimum marker count
        markers_to_keep = X.columns[
            (X != 0).any(axis=0) & (X.nunique() > 1) & (X.sum() >= min_marker_count)
        ]

        transform_func = np.log1p if transformation == "log1p" else lambda x: x
        X_perm = [
            transform_func(x)
            for x in permutations(X, n=n_permutations, random_seed=random_seed)
        ]
        X = transform_func(X)

        # Apply marker filter after transformation
        X = X.loc[:, markers_to_keep]
        X_perm = [x.loc[:, markers_to_keep] for x in X_perm]

        _compute_morans_i_per_marker = partial(
            _compute_morans_i,
            marker_count_matrix=X,
            marker_count_matrix_permuted=X_perm,
            weight_matrix=W,
            normalization_factor=C,
        )
        markers = X.columns

        results = pd.DataFrame(
            map(_compute_morans_i_per_marker, markers),
            columns=[
                "marker",
                "morans_i",
                "morans_z",
                "perm_mean",
                "perm_std",
                "morans_p_value",
            ],
        )
        return results

    start_time = time.perf_counter()
    logger.info("Computing Moran's I for component: %s", component_id)

    results = _compute_morans_on_current_graph()
    results["component"] = component_id

    run_time = time.perf_counter() - start_time
    logger.info(
        "Finished computing Moran's I for component: %s in %.2fs",
        component_id,
        run_time,
    )
    results = results.drop(columns=["perm_mean", "perm_std"])
    return results


def polarization_scores_component_df(
    component_id: str,
    component_df: pd.DataFrame,
    use_full_bipartite: bool,
    transformation: PolarizationTransformationTypes = "log1p",
    n_permutations: int = 50,
    min_marker_count: int = 2,
    random_seed: int | None = None,
):
    """Calculate Moran's I statistics for a component.

    See `polarization_scores_component_graph` for details.

    :param component_id: the id of the component
    :param component_df: A data frame with an edgelist for a single connected component
    :param use_full_bipartite: use the bipartite graph instead of the projection (UPIA)
    :param transformation: the count transformation method to use (raw, log1p)
    :param n_permutations: the number of permutations to use to estimate the
                           null-hypothesis for the Moran's I statistic
    :param min_marker_count: the minimum number of counts of a marker to calculate
                             the Moran's I statistic
    :param random_seed: the random seed to use to ensure that the permutations
                        are reproducible across runs
    :returns: a pd.DataFrame with the polarization statistics for each antibody
    :rtype: pd.DataFrame
    :raises: AssertionError when the input is not valid
    """
    graph = Graph.from_edgelist(
        edgelist=component_df,
        add_marker_counts=True,
        simplify=True,
        use_full_bipartite=use_full_bipartite,
    )

    component_result = polarization_scores_component_graph(
        graph=graph,
        component_id=component_id,
        transformation=transformation,
        n_permutations=n_permutations,
        min_marker_count=min_marker_count,
        random_seed=random_seed,
    )
    return component_result


def polarization_scores(
    edgelist: pd.DataFrame,
    use_full_bipartite: bool = False,
    transformation: PolarizationTransformationTypes = "log1p",
    n_permutations: int = 0,
    min_marker_count: int = 2,
    random_seed: int | None = None,
) -> pd.DataFrame:
    """Calculate Moran's I statistics for an edgelist.

    Compute polarization scores from an `edgelist`.
    The function iterates all the components to compute polarization scores
    for each antibody in the component. The scores are computed using Moran's
    I autocorrelation to measure how clustered/localised the spatial patterns of
    the antibody is in the component's graph. Spatial weights (`w`) are derived
    directly from the graph. The function returns a pd.DataFrame with the following
    columns:
      morans_i, morans_p_value, morans_z, morans_p_adjusted, marker, component
      (morans_p_value_sim and morans_z_sim if `permutations` > 0)
    :param edgelist: an edge list (pd.DataFrame) with a component column
    :param use_full_bipartite: use the bipartite graph instead of the projection (UPIA)
    :param transformation: the count transformation method to use (raw, log1p)
    :param n_permutations: the number of permutations for simulated Z-score (z_sim)
                           estimation (if n_permutations>0)
    :param min_marker_count: the minimum number of counts of a marker to calculate
                             the Moran's I statistic
    :param random_seed: the random seed to use for reproducibility
    :returns: a pd.DataFrames with all the polarization scores
    :rtype: pd.DataFrame
    :raises: AssertionError when the input is not valid
    """
    if transformation not in get_args(PolarizationTransformationTypes):
        raise AssertionError(
            f"incorrect value for count transformation {transformation}"
        )

    if "component" not in edgelist.columns:
        raise AssertionError("Edge list is missing the component column")

    if n_permutations < 0:
        logger.warning(
            "Setting `n_permutations < 0` will mean no z-scores and p-values will be calculated."
        )

    logger.debug(
        "Computing polarization for edge list with %i elements",
        edgelist.shape[0],
    )

    def data():
        with get_pool_executor() as pool:
            polarization_function = partial(
                polarization_scores_component_df,
                use_full_bipartite=use_full_bipartite,
                transformation=transformation,
                n_permutations=n_permutations,
                min_marker_count=min_marker_count,
                random_seed=random_seed,
            )
            yield from pool.starmap(
                polarization_function,
                (
                    (component_id, component_df)
                    for component_id, component_df in edgelist.groupby(
                        "component", observed=True
                    )
                ),
                chunksize=10,
            )

    # create dataframe with all the scores
    logger.debug("Concatenating the polarization dataframes")
    scores = pd.concat(data(), axis=0)
    if scores.empty:
        logger.warning("Polarization results were empty")
        return scores

    scores.insert(
        scores.columns.get_loc("morans_p_value") + 1,
        "morans_p_adjusted",
        correct_pvalues(scores["morans_p_value"].to_numpy()),
    )

    logger.debug("Polarization scores for edge list computed")
    return scores


class PolarizationAnalysis(PerComponentAnalysis):
    """Run polarization analysis on each component."""

    ANALYSIS_NAME = "polarization"

    def __init__(
        self,
        transformation_type: PolarizationTransformationTypes,
        n_permutations: int,
        min_marker_count: int,
        random_seed: int | None = None,
    ):
        """Initialize polarization analysis.

        :param transformation: the count transformation method to use (raw, log1p)
        :param n_permutations: the number of permutations to use to estimate the
                               null-hypothesis for the Moran's I statistic
        :param min_marker_count: the minimum number of counts of a marker to calculate
                                 the Moran's I statistic
        :param random_seed: set a random seed to ensure reproducibility when calculating z-scores
                            and p-values.
        """
        if transformation_type not in get_args(PolarizationTransformationTypes):
            raise AssertionError(
                f"incorrect value for count transformation {transformation_type}"
            )
        self.transformation = transformation_type
        self.permutations = n_permutations
        self.min_marker_count = min_marker_count
        self.random_seed = random_seed

    def run_on_component(self, component: Graph, component_id: str) -> pd.DataFrame:
        """Run polarization analysis on component."""
        logger.debug("Running polarization analysis on component %s", component_id)
        return polarization_scores_component_graph(
            graph=component,
            component_id=component_id,
            transformation=self.transformation,
            n_permutations=self.permutations,
            min_marker_count=self.min_marker_count,
            random_seed=self.random_seed,
        )

    def post_process_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Post process polarization analysis data.

        This will adjust the calculated p-values for the calculate Moran's I statistics.
        """
        logger.debug("Post processing polarization analysis data")
        if data.empty:
            return data
        data.insert(
            data.columns.get_loc("morans_p_value") + 1,
            "morans_p_adjusted",
            correct_pvalues(data["morans_p_value"].to_numpy()),
        )
        return data

    def add_to_pixel_dataset(
        self, data: pd.DataFrame, pxl_dataset: PixelDataset
    ) -> PixelDataset:
        """Add data to the polarization field of the PixelDataset."""
        logger.debug("Adding polarization analysis data to PixelDataset")
        pxl_dataset.polarization = data
        return pxl_dataset


def get_differential_polarity(
    polarity_data: pd.DataFrame,
    reference: str,
    targets: str | list[str] | None = None,
    contrast_column: str = "sample",
    value_column: str = "morans_z",
) -> pd.DataFrame:
    """Calculate the differential polarity.

    :param polarity_data: The polarity data frame.
    :param target: The label for target components in the contrast_column.
    :param reference: The label for reference components in the contrast_column.
    :param contrast_column: The column to use for the contrast. Defaults to "sample".
    :param value_column: What polarity metric to use. Defaults to "morans_z".

    :return: The differential polarity.
    :rtype: pd.DataFrame
    """
    if targets is None:
        targets = polarity_data[contrast_column].unique()
        targets = list(set(targets) - {reference})
    elif isinstance(targets, str):
        targets = [targets]

    if len(targets) > 5:
        logger.warning(
            "There are more than 5 targets in the dataset. This may take a while."
        )
    merged_differential_polarity = pd.DataFrame()
    for target in targets:
        differential_polarity = (
            polarity_data.groupby("marker")
            .apply(
                lambda marker_data: wilcoxon_test(
                    marker_data,
                    reference=reference,
                    target=target,
                    contrast_column=contrast_column,
                    value_column=value_column,
                )
            )
            .reset_index()
        )

        # If a marker appears only in one of the datasets,
        # it's differential value will be NAN
        nan_values = differential_polarity[
            differential_polarity["median_difference"].isna()
        ].index
        differential_polarity.drop(
            nan_values,
            axis="index",
            inplace=True,
        )

        _, pvals_corrected, *_ = multipletests(
            differential_polarity["p_value"], method="bonferroni"
        )
        differential_polarity["p_adj"] = pvals_corrected
        differential_polarity["target"] = target
        merged_differential_polarity = pd.concat(
            (merged_differential_polarity, differential_polarity), axis=0
        )

    return merged_differential_polarity



"""MPX configuration.

Copyright © 2025 Pixelgen Technologies AB.
"""

from pixelator.common.config import get_position_in_parent, load_antibody_panel
from pixelator.mpx.config.config_instance import config

__all__ = [
    "config",
    "get_position_in_parent",
    "load_antibody_panel",
]



"""Module for utilities for working with pixeldatasets.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
import os
import warnings
from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd
from anndata import AnnData, ImplicitModificationWarning, read_h5ad
from graspologic_native import leiden

from pixelator.common.statistics import (
    clr_transformation,
    log1p_transformation,
)
from pixelator.common.types import PathType
from pixelator.mpx.graph import components_metrics
from pixelator.mpx.graph.constants import (
    LEIDEN_RESOLUTION,
    RELATIVE_ANNOTATE_RESOLUTION,
)

if TYPE_CHECKING:
    from pixelator.common.config import AntibodyPanel

logger = logging.getLogger(__name__)


def update_metrics_anndata(adata: AnnData, inplace: bool = True) -> Optional[AnnData]:
    """Update any metrics in the AnnData instance.

    This will  update the QC metrics (`var` and `obs`) of
    the AnnData object given as input. This function is typically used
    when the AnnData object has been filtered and one wants the QC metrics
    to be updated accordingly.

    :param adata: an AnnData object
    :param inplace: If `True` performs the operation inplace
    :returns: the updated AnnData object or None if inplace is True
    :rtype: Optional[AnnData]
    """
    logger.debug(
        "Updating metrics in AnnData object with %i components and %i markers",
        adata.n_obs,
        adata.n_vars,
    )

    if not inplace:
        adata = adata.copy()

    df = adata.to_df()

    # update the var layer (antibody metrics)
    # we ignore the warning here, since we actually want to force and update of the
    # `adata.var` data frame.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ImplicitModificationWarning)
        adata.var["antibody_count"] = df.sum().astype(int)
    adata.var["components"] = (df != 0).sum()
    adata.var["antibody_pct"] = (
        adata.var["antibody_count"] / adata.var["antibody_count"].sum()
    )

    # update the obs layer (components metrics)
    adata.obs["antibodies"] = np.sum(adata.X > 0, axis=1)

    logger.debug("Metrics in AnnData object updated")
    return None if inplace else adata


def _enforce_edgelist_types(edgelist: pd.DataFrame) -> pd.DataFrame:
    """Enforce the data types of the edgelist."""
    # Enforcing the types of the edgelist reduces the memory
    # usage by roughly 2/3s.

    required_types = {
        "count": "uint16",
        "upia": "category",
        "upib": "category",
        "umi": "category",
        "marker": "category",
        "sequence": "category",
        "component": "category",
    }

    # if the dataframe is empty just enforce the types.
    if edgelist.shape[0] == 0:
        edgelist = pd.DataFrame(columns=required_types.keys())

    # If we have the optional sample column, this should be
    # set to use a categorical type
    if "sample" in edgelist.columns:
        required_types["sample"] = "category"

    # If all of the prescribed types are already set, just return the edgelist
    type_dict = edgelist.dtypes.to_dict()
    if all(type_dict[key] == type_ for key, type_ in required_types.items()):
        return edgelist

    return edgelist.astype(
        required_types,
        # Do not copy here, since otherwise the memory usage
        # blows up
        copy=False,
    )


def antibody_metrics(edgelist: pd.DataFrame) -> pd.DataFrame:
    """Calculate metrics for each antibody/marker.

    A helper function that computes a dataframe of antibody
    metrics for each antibody (marker) present in the edge list
    given as input. The metrics include: total count, relative
    count and the number of components where the antibody is detected.

    :param edgelist: an edge list dataframe with a membership column
    :returns: a pd.DataFrame with the antibody metrics per antibody
    :rtype: pd.DataFrame
    :raises: AssertionError when the input edge list is not valid
    """
    if "component" not in edgelist.columns:
        raise AssertionError("Edge list is missing the component column")

    logger.debug(
        "Computing antibody metrics for dataset with %i elements", edgelist.shape[0]
    )

    # compute metrics
    antibody_metrics = (
        edgelist.groupby("marker", observed=True)
        .agg(
            {
                "count": "sum",
                "component": "nunique",
            }
        )
        .astype(int)
    )
    antibody_metrics.columns = [  # type: ignore
        "antibody_count",
        "components",
    ]

    # add relative counts
    antibody_metrics["antibody_pct"] = (
        antibody_metrics["antibody_count"] / antibody_metrics["antibody_count"].sum()
    ).astype(float)

    logger.debug("Antibody metrics computed")
    return antibody_metrics


def component_antibody_counts(edgelist: pd.DataFrame) -> pd.DataFrame:
    """Calculate antibody counts per component.

    A helper function that computes a dataframe of antibody
    counts for each component present in the edge list given
    as input (component column).

    :param edgelist: an edge list dataframe with a membership column
    :returns: a pd.DataFrame with the antibody counts per component
    :rtype: pd.DataFrame
    :raises: AssertionError when the input edge list is not valid
    """
    if "component" not in edgelist.columns:
        raise AssertionError("Edge list is missing the component column")

    logger.debug(
        "Computing antibody counts for edge list with %i edges and %i markers",
        edgelist.shape[0],
        edgelist.shape[1],
    )

    # iterate the components to obtain the metrics of each component
    # TODO this seems to be memory demanding so a simpler groupby() over
    # the component column may perform better in terms of memory
    df = (
        edgelist.groupby(["component", "marker"], observed=True)
        .agg("size")
        .unstack()
        .fillna(0)
    ).astype(int)
    df.index.name = "component"

    logger.debug("Antibody counts computed")
    return df


def read_anndata(filename: str) -> AnnData:
    """Read an AnnData object from a h5ad file.

    A simple wrapper to read/parse AnnData (h5ad) files.

    :param filename: the path to the AnnData file (h5ad)
    :returns: an AnnData object
    :rtype: AnnData
    :raises: AssertionError when the input is not valid
    """
    if not os.path.isfile(filename):
        raise AssertionError(f"input {filename} does not exist")
    if not filename.endswith("h5ad"):
        raise AssertionError(f"input {filename} has a wrong extension")
    return read_h5ad(filename=filename)


def write_anndata(adata: AnnData, filename: PathType) -> None:
    """Write anndata instance to file.

    A simple wrapper to write/save an AnnData object to a file.

    :param adata: the AnnData object to be saved
    :param filename: the path to save AnnData file (h5ad)
    :returns: None
    :rtype: None
    """
    adata.write(filename=filename, compression="gzip")


def _compute_sub_communities(
    component_edgelist: pd.DataFrame, n_edges_reconnect: int | None = None
) -> pd.Series:
    component_edgelist = (
        component_edgelist.groupby(["upia", "upib"], observed=True)["count"]
        .count()
        .reset_index()
        .sort_values(["upia", "upib"])
    )
    edgelist_tuple = list(
        map(tuple, np.array(component_edgelist[["upia", "upib", "count"]]))
    )
    _, component_communities_dict = leiden(
        edgelist_tuple,
        resolution=RELATIVE_ANNOTATE_RESOLUTION * LEIDEN_RESOLUTION,
        seed=42,
        # These parameters are used to sync up the native implementation with
        # the python implementation we originally used.
        use_modularity=True,
        iterations=1,
        randomness=0.001,
        trials=1,
        starting_communities=None,
    )
    component_communities = pd.Series(component_communities_dict)

    return component_communities


def _assess_doublet(component_edgelist: pd.DataFrame) -> tuple[bool, int]:
    """Check whether a component is a potential doublet and how many edges should be removed to split it.

    A component is a potential doublet if a) it has more than one community and
    b) the second largest community is at least 20% of the size of the largest
    community. A lower resolution is to be used for annotation of potential doublets
    compared to the component recovery in the graph phase. The reduction factor in
    annotate resolution is set by RELATIVE_ANNOTATE_RESOLUTION (default is 0.5).

    """
    component_communities = _compute_sub_communities(component_edgelist)
    component_community_sizes = component_communities.value_counts().sort_values(
        ascending=False
    )
    if len(component_community_sizes) > 1 and component_community_sizes.iloc[1] > (
        0.2 * component_community_sizes.iloc[0]
    ):
        edges_to_remove = (
            component_edgelist["upia"].map(component_communities)
            != component_edgelist["upib"].map(component_communities)
        ).sum()
        return True, edges_to_remove
    else:
        return False, 0


def mark_potential_doublets(
    edgelist: pd.DataFrame,
) -> tuple[pd.Series, pd.Series]:
    """Mark whether a component is a potential doublet.

    A component is a potential doublet if a) it has more than one community and
    b) the second largest community is at least 20% of the size of the largest
    community.

    :param edgelist: the edge list dataframe containing component labels.
    :returns: a boolean series indicating whether a component is a potential doublet.
    :rtype: pd.Series
    """
    is_potential_doublet = pd.Series(index=edgelist["component"].unique(), dtype=bool)
    n_edges_to_split_doublet = pd.Series(
        index=edgelist["component"].unique(), dtype=int
    )
    for component_id, component_edgelist in edgelist.groupby("component"):
        is_potential_doublet[component_id], n_edges_to_split_doublet[component_id] = (
            _assess_doublet(component_edgelist)
        )

    return is_potential_doublet, n_edges_to_split_doublet


def edgelist_to_anndata(
    edgelist: pd.DataFrame,
    panel: AntibodyPanel,
) -> AnnData:
    """Convert an edgelist to an anndata object.

    A helper function to build an AnnData object from an edge list (dataframe).
    The `panel` will be used to add extra information (`var` layer) and to ensure
    that all the antibodies are included in the AnnData object.

    The AnnData will have the following layers:

    .X = the component to antibody counts
    .var = the antibody metrics
    .obs = the component metrics
    .obsm["clr"] = the transformed (clr) component to antibody counts
    .obsm["log1p"] = the transformed (log1p) component to antibody counts

    :param edgelist: an edge list (pd.DataFrame)
    :param panel: the AntibodyPanel of the panel used to generate the data
    :returns: an AnnData object
    :rtype: AnnData
    """
    logger.debug("Creating AnnData from edge list with %i edges", edgelist.shape[0])

    missing_markers = ",".join(
        set(panel.markers).difference(set(edgelist["marker"].unique()))
    )
    if missing_markers:
        msg = (
            "The given 'panel' is missing markers "
            f"({missing_markers}) in the edge list, "
            "these will be added with 0 counts"
        )
        logger.warning(msg)

    # compute antibody counts and re-index
    counts_df = component_antibody_counts(edgelist=edgelist)
    counts_df = counts_df.reindex(columns=panel.markers, fill_value=0)
    counts_df.index = counts_df.index.astype(str)
    counts_df.columns = counts_df.columns.astype(str)

    # compute components metrics (obs) and re-index
    components_metrics_df = components_metrics(edgelist=edgelist)
    components_metrics_df = components_metrics_df.reindex(index=counts_df.index)
    (
        components_metrics_df["is_potential_doublet"],
        components_metrics_df["n_edges_to_split_doublet"],
    ) = mark_potential_doublets(edgelist=edgelist)
    # compute antibody metrics (var) and re-index
    antibody_metrics_df = antibody_metrics(edgelist=edgelist)
    antibody_metrics_df = antibody_metrics_df.reindex(index=panel.markers, fill_value=0)
    # Do a dtype conversion of the columns here since AnnData cannot handle
    # a pyarrow arrays.
    antibody_metrics_df = antibody_metrics_df.astype(
        {"antibody_count": "int64", "antibody_pct": "float32"}
    )

    # create AnnData object
    adata = AnnData(
        X=counts_df,
        obs=components_metrics_df,
        var=antibody_metrics_df,
    )

    # add extra panel variables to var
    adata.var["nuclear"] = panel.df["nuclear"].to_numpy()
    adata.var["control"] = panel.df["control"].to_numpy()

    # add normalization layers
    counts_df_clr = clr_transformation(df=counts_df, axis=1)
    counts_df_log1p = log1p_transformation(df=counts_df)
    adata.obsm["clr"] = counts_df_clr
    adata.obsm["log1p"] = counts_df_log1p

    logger.debug("AnnData created")
    return adata



"""Top level module for analysis subpackage of pixelator.pna.

Copyright © 2024 Pixelgen Technologies AB.
"""

from pixelator.pna.analysis.proximity import calculate_differential_proximity

__all__ = ["calculate_differential_proximity"]



"""Pixelator for use with MPX data.

Copyright © 2022 Pixelgen Technologies AB.
"""

# Adding imports here as shortcuts to be able to import like
# import pixelator as mpx
# mpx.read("<file path>")
# and similar
from pixelator.mpx.pixeldataset import read, PixelDataset  # noqa
from pixelator.mpx.pixeldataset.aggregation import simple_aggregate  # noqa


__all__ = ["read", "simple_aggregate", "PixelDataset"]



"""Module that contains functions for demultiplex reads belonging to antibody panels.

Copyright © 2022 Pixelgen Technologies AB.
"""

import json
import logging
import subprocess
from pathlib import Path
from subprocess import CalledProcessError

# List is used as type hint in comment
from typing import List  # noqa: F401

from pixelator.common.config.panel import AntibodyPanel

logger = logging.getLogger(__name__)

DEMUX_PERCENTAGE_THRESHOLD = 0.5


def demux_fastq(
    input: str,
    output: str,
    failed: str,
    report: str,
    panel: AntibodyPanel,
    mismatches: float,
    barcodes: str,
    min_length: int,
    cores: int,
    verbose: bool,
    sample_id: str,
) -> bool:
    """Demultiplex the input fastq file into separate files based on the barcodes.

    This function is a wrapper around `cutadapt` to process a `fastq`
    file with molecular pixelation data. A `fasta` file with the barcodes
    to use to demultiplex must be provided (antibody name and its sequence).
    The fastq will be processed to demultiplex its barcode with the ones
    provided in the fasta file (--barcodes). One fastq file per barcode
    will be generated (containing the reads that matched). Another single
    file with the reads that failed to demultiplex will also be created.

    :param input: the path to the fastq file (must contain the barcode)
    :param output: the path to the output file (processed)
    :param failed: the path to the failed file (discarded)
    :param report: the path to the json report
    :param panel: the path to the panel used for demultiplexing
    :param mismatches: the number of mismatches allowed (a percentage)
    :param barcodes: the fasta file containing the barcodes (reference)
    :param min_length: the minimum overlap required in the barcode
    :param cores: the number of cores to use
    :param verbose: run in verbose mode when true
    :param sample_id: the sample id
    :returns: true if demux results were ok
    :raises ValueError: raises an exception
            OSError: raises an exception
            CalledProcessError: raises an exception
            RuntimeError: raises an exception
    """
    args = [
        "cutadapt",
        "-e",
        str(mismatches),
        "--adapter",
        f"file:{barcodes}",
        "--cores",
        str(cores),
        "--action=none",
        "--no-indels",
        "--untrimmed-output",
        failed,
        "--overlap",
        str(min_length),
        "--json",
        report,
        "--output",
        output,
        input,
    ]  # type: List[str]

    if verbose:
        args += ["--debug"]

    logger.debug("Invoking cutadapt: %s", " ".join(args))

    try:
        proc = subprocess.Popen(  # type: ignore
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            shell=False,
        )
        (stdout, errmsg) = proc.communicate()
    except ValueError as exc:
        logger.error("ERROR running cutadapt. Incorrect arguments")
        raise exc
    except OSError as exc:
        logger.error("ERROR running cutadapt. Executable not found")
        raise exc
    except CalledProcessError as exc:
        logger.error("ERROR running cutadapt. Program returned error")
        raise exc

    # check output file exists
    if not Path(report).is_file():
        error = (
            f"ERROR running cutadapt.\nOutput file not present "
            f"{report}\n{stdout.decode('utf-8')}\n{errmsg.decode('utf-8')}"
        )
        logger.error(error)
        raise RuntimeError(error)

    modified_data = {}
    with open(report, "r") as fh:
        file_data = json.load(fh)

        for adapters in file_data["adapters_read1"]:
            adapters["name"] = panel.get_marker_id(adapters["name"])

        modified_data = file_data
        modified_data["sample_id"] = sample_id

    with open(report, "w") as fh:
        json.dump(modified_data, fh, indent=4)

    return check_demux_results_are_ok(modified_data, sample_id)


def check_demux_results_are_ok(report_data: dict, sample_id: str) -> bool:
    """Check if the demultiplexing results are ok."""
    read_counts = report_data["read_counts"]
    input_reads = read_counts["input"]
    output_reads = read_counts["output"]

    if output_reads / input_reads < DEMUX_PERCENTAGE_THRESHOLD:
        logger.error(
            (
                "Low demultiplexing rate (%.2f%%) for %s. Your results may not be reliable."
                "Are you sure you have used the correct panel file?"
            ),
            output_reads / input_reads * 100,
            sample_id,
        )
        return False
    return True



"""Demultiplex reads based on the PNA assay design.

Copyright © 2024 Pixelgen Technologies AB.
"""

import abc
import heapq
import math
import typing
from collections import Counter, defaultdict
from typing import Literal, Sequence, overload

import numpy as np
import numpy.typing as npt
import pyarrow as pa
from dnaio import SequenceRecord

from pixelator.common.exceptions import PixelatorBaseException
from pixelator.pna.config import PNAAntibodyPanel, PNAAssay, get_position_in_parent


def create_barcode_group_to_batch_mapping(
    group_sizes: dict[tuple[str, str], int],
    reads_per_chunk: int = 50_000_000,
    max_chunks: int = 8,
) -> dict[tuple[str, str], int]:
    """Create groups of (PID1, PID2) pairs.

    The input mapping is a dictionary of (PID1, PID2) pairs to the number of
    reads that have been found to contain that pair of PID barcodes.

    The total number of reads in a supergroup can be less or more than the target_chunk_size.

    :param group_sizes: the mapping of (PID1, PID2) pairs to the number of reads
    :param reads_per_chunk: the target number of reads per chunk
    :param max_chunks: the maximum number of groups
    """
    # We can initialize the group mapping already with a default batch index of 0

    marker_counts: typing.Counter[tuple[str, str]] = Counter()

    for keys, v in group_sizes.items():
        marker_counts[keys] += v

    target_chunks = int(
        min(
            math.ceil(sum(marker_counts.values()) / reads_per_chunk),
            max_chunks,
        )
    )

    marker_labels = partition_greedy(list(marker_counts.values()), target_chunks)
    marker_group_mapping = {
        marker: int(group) for marker, group in zip(marker_counts.keys(), marker_labels)
    }

    return marker_group_mapping


def partition_greedy(items: Sequence[int], n: int) -> npt.NDArray[np.int32]:
    """Greedy number partitioning of a set of items with a given weight into n subsets.

    Each subset has an approximately equal sum of weights.

    :param counts: a dictionary of items and their weights
    :param n: the number of subsets
    """
    # Initialize a priority queue with n empty subsets
    subset_sums = [(0, i) for i in range(n)]
    heapq.heapify(subset_sums)

    labels = np.zeros(len(items), dtype=np.int32)

    for idx, count in enumerate(sorted(items, reverse=True)):
        # Pop the subset with the smallest sum
        current_sum, subset_index = heapq.heappop(subset_sums)
        # Assign the item to this subset
        labels[idx] = subset_index
        # Push the updated subset back into the priority queue
        heapq.heappush(subset_sums, (current_sum + count, subset_index))

    return labels


K = typing.TypeVar("K", bound=tuple[str, str] | str)


def independent_marker_groups_mapping(
    group_sizes: dict[tuple[str, str], int],
    reads_per_chunk: int = 50_000_000,
    max_chunks: int = 8,
) -> tuple[dict[str, int], dict[str, int]]:
    """Combine grouped reads by markers into approximately equal sized chunks.

    This is the classical partitioning problem.
    We use the simple greedy number partitioning algorithm here.

    The number of groups is determined by the `target_chunk_reads`.
    The number of groups will be capped by `max_chunk_count` should the calculated number of groups
    exceed this value.

    Params:
        group_sizes: the mapping of (marker1, marker2) pairs to the number of reads
        target_chunk_count: the number of chunks to partition the markers into

    Returns:
        A tuple with a dict for marker1 and marker2. The dict map the marker to the group index.

    """
    marker1_counts: typing.Counter[str] = Counter()
    marker2_counts: typing.Counter[str] = Counter()

    for (m1, m2), v in group_sizes.items():
        marker1_counts[m1] += v
        marker2_counts[m2] += v

    target_m1_chunks = int(
        min(
            math.ceil(sum(marker1_counts.values()) / reads_per_chunk),
            max_chunks,
        )
    )
    target_m2_chunks = int(
        min(
            math.ceil(sum(marker2_counts.values()) / reads_per_chunk),
            max_chunks,
        )
    )

    marker1_labels = partition_greedy(list(marker1_counts.values()), target_m1_chunks)
    marker1_group_mapping = {
        marker: int(group)
        for marker, group in zip(marker1_counts.keys(), marker1_labels)
    }

    offset = np.max(marker1_labels) + 1
    marker2_labels = partition_greedy(list(marker2_counts.values()), target_m2_chunks)
    marker2_group_mapping = {
        marker: int(offset + group)
        for marker, group in zip(marker2_counts.keys(), marker2_labels)
    }

    return marker1_group_mapping, marker2_group_mapping


class DemuxRecordBatch:
    """A batch of demux records.

    Batches are used to collect records before they are serialized and sent to the writer process.
    A batch uses a pre-allocated fixed length numpy array per field to store the records.

    :params batch_size: the maximum number of records in the batch
    """

    # TODO: Make this modular by using the assay design
    #   the packed array lengths for umi's and uei are hard-coded now
    _record_batch_schema = pa.schema(
        [
            pa.field("marker_1", pa.uint16()),
            pa.field("marker_2", pa.uint16()),
            pa.field("molecule", pa.binary(32)),
        ]
    )

    def __init__(self, capacity=10_000):
        """Initialize the DemuxRecordBatch object.

        :param capacity: the maximum number of records in the batch
        """
        self._batch_size = capacity
        self._size = 0
        self.marker1 = np.zeros(self._batch_size, dtype=np.uint16)
        self.marker2 = np.zeros(self._batch_size, dtype=np.uint16)
        self.molecule = np.zeros(self._batch_size, dtype="V32")

    @classmethod
    def schema(cls) -> pa.Schema:
        """Return the schema of the record batch returned by `to_arrow`."""
        return cls._record_batch_schema

    def add_record(self, marker1: int, marker2: int, molecule: bytes):
        """Add a new record to the batch.

        :param marker1: the first marker index
        :param marker2: the second marker index
        :param molecule: the molecule embedding
        """
        _size = self._size

        if _size < self._batch_size:
            self.marker1[_size] = marker1
            self.marker2[_size] = marker2
            self.molecule[_size] = molecule
            self._size += 1
        else:
            raise IndexError("Batch is full")

    def __len__(self) -> int:
        """Return the current number of records in the batch."""
        return self._size

    def capacity(self) -> int:
        """Return the maximum number of records in the batch."""
        return self._batch_size

    def clear(self):
        """Clear the batch.

        Note that the memory is not released, only the size is reset to 0.
        """
        self._size = 0

    def _verify_m1_markers(self, allowed_ids: Sequence[int]) -> bool:
        """Check if all markers are in the allowed set."""
        mask = np.isin(self.marker1[: self._size], allowed_ids)
        return bool(np.all(mask))

    def _verify_m2_markers(self, allowed_ids: Sequence[int]) -> bool:
        """Check if all markers are in the allowed set."""
        mask = np.isin(self.marker2[: self._size], allowed_ids)
        return bool(np.all(mask))

    def to_arrow(self) -> pa.RecordBatch:
        """Convert a number of cached records to a pyarrow RecordBatch."""
        _size = self._size
        # Passing the size option to pa.array does not seem to work for some reason.
        # So we are using slices instead.
        return pa.RecordBatch.from_arrays(
            [
                pa.array(
                    self.marker1[:_size],
                    type=self._record_batch_schema.field("marker_1").type,
                ),
                pa.array(
                    self.marker2[:_size],
                    type=self._record_batch_schema.field("marker_2").type,
                ),
                pa.array(
                    self.molecule[:_size],
                    type=self._record_batch_schema.field("molecule").type,
                ),
            ],
            schema=self._record_batch_schema,
        )

    def serialize(self) -> pa.Buffer:
        """Serialize the record batch as Arrow IPC message."""
        return self.to_arrow().serialize()


class BarcodeDemuxingError(PixelatorBaseException):
    """An error occurred during barcode demultiplexing."""

    pass


_ENCODING_LUT = np.zeros((256, 3), dtype=np.uint8)

# Numpy array representation is in most to least significant bit order

_ENCODING_LUT[ord("A")] = np.array([0, 1, 1], dtype=np.uint8)  # b'110'
_ENCODING_LUT[ord("C")] = np.array([1, 0, 1], dtype=np.uint8)  # b'101'
_ENCODING_LUT[ord("T")] = np.array([1, 1, 0], dtype=np.uint8)  # b'011'
_ENCODING_LUT[ord("G")] = np.array([0, 0, 0], dtype=np.uint8)  # b'000'


_ENCODING_LUT_2BIT = np.zeros((256, 2), dtype=np.uint8)

_ENCODING_LUT_2BIT[ord("A")] = np.array([0, 0], dtype=np.uint8)  # b'00'
_ENCODING_LUT_2BIT[ord("C")] = np.array([1, 0], dtype=np.uint8)  # b'01'
_ENCODING_LUT_2BIT[ord("G")] = np.array([0, 1], dtype=np.uint8)  # b'10'
_ENCODING_LUT_2BIT[ord("T")] = np.array([1, 1], dtype=np.uint8)  # b'11'

_DECODING_LUT = np.zeros((256), dtype=np.uint8)
_DECODING_LUT[6] = ord(b"A")
_DECODING_LUT[5] = ord(b"C")
_DECODING_LUT[3] = ord(b"T")
_DECODING_LUT[0] = ord(b"G")
_DECODING_LUT[7] = ord(b"N")

_RECODING_3BIT_TO_2BIT_LUT = np.zeros((8, 2), dtype=np.uint8)

_RECODING_3BIT_TO_2BIT_LUT[6] = np.array([0, 0], dtype=np.uint8)  # A -> b'00'
_RECODING_3BIT_TO_2BIT_LUT[5] = np.array([1, 0], dtype=np.uint8)  # C -> b'01'
_RECODING_3BIT_TO_2BIT_LUT[0] = np.array([0, 1], dtype=np.uint8)  # G -> b'01'
_RECODING_3BIT_TO_2BIT_LUT[3] = np.array([1, 1], dtype=np.uint8)  # T -> b'11'


class PNAEmbedding:
    """A class for constructing a compact binary embedding of a PNA amplicon reads.

    The embedding assumes a maximum length of 32 nucleotides
    for each UMI and 15 nucleotides for the UEI.
    """

    def __init__(self, assay: PNAAssay):
        """Initialize the PNAEmbedding object.

        :param assay: the assay design
        """
        self.assay = assay

        self._umi1_len = assay.get_region_by_id("umi-1").max_len
        self._umi2_len = assay.get_region_by_id("umi-2").max_len
        self._uei_len = assay.get_region_by_id("uei").max_len

        if self._umi1_len > 32:
            raise RuntimeError("UMI1s longer than 32 nucleotides are not supported")
        if self._umi2_len > 32:
            raise RuntimeError("UMI2s longer than 32 nucleotides are not supported")
        if self._uei_len > 21:
            raise RuntimeError("UEIs longer than 21 nucleotides are not supported")

        self._umi1_slice = slice(0, self._umi1_len * 3)
        self._umi2_slice = slice(96, 96 + self._umi2_len * 3)
        self._uei_slice = slice(192, 192 + self._uei_len * 3)

    def get_umi1_bytes(self, embedded: bytes):
        """Return the bytes of the UMI1 from the molecule embedding."""
        return embedded[0:12]

    def get_umi2_bytes(self, embedded: bytes):
        """Return the bytes of the UMI2 from the molecule embedding."""
        return embedded[12:24]

    def get_uei_bytes(self, embedded: bytes):
        """Return the bytes of the UEI from the molecule embedding."""
        return embedded[24:32]

    def _encode_umi1(self, umi1: bytes, output):
        assert len(umi1) == self._umi1_len

        umi1_arr = np.frombuffer(umi1, dtype=np.uint8, count=len(umi1))
        umi1_vec_slice = output[self._umi1_slice].reshape((self._umi1_len, 3))
        np.take(_ENCODING_LUT, umi1_arr, axis=0, out=umi1_vec_slice)

    def _encode_umi2(self, umi2: bytes, output):
        assert len(umi2) == self._umi2_len

        umi1_arr = np.frombuffer(umi2, dtype=np.uint8, count=len(umi2))
        umi1_vec_slice = output[self._umi2_slice].reshape((self._umi1_len, 3))
        np.take(_ENCODING_LUT, umi1_arr, axis=0, out=umi1_vec_slice)

    def _encode_uei(self, uei: bytes, output):
        assert len(uei) == self._uei_len

        uei_arr = np.frombuffer(uei, dtype=np.uint8, count=len(uei))

        # Write all ones if the UEI is not present
        # Otherwise encode the UEI into the last 8 bytes
        # 15 nucleotides * 3 bits = 45 bits
        # padded to 64 bits = 8 bytes
        if np.any(uei_arr == ord("N")):
            output[self._uei_slice] = 1
        else:
            uei_vec_slice = output[self._uei_slice].reshape((15, 3))
            np.take(_ENCODING_LUT, uei_arr, axis=0, out=uei_vec_slice)

    def encode(self, umi1: bytes, umi2: bytes, uei: bytes):
        """Pack the unique molecule identifiers into a single 256-bit vector.

        Each nucleotide is packed into 3 bits, so 8 nucleotides per 24-bit
        Padding bits are added to make the total length 256 bits and to separate
        the regions on byte boundaries.

        256 bits is a common length for SIMD instructions (eg. 1 AVX-2 register or 2 Arm NEON registers)
        and thus commonly used for efficient similarity search. (eg. in FAISS).
        """
        vec = np.zeros(256, dtype=np.uint8)

        self._encode_umi1(umi1, vec)
        self._encode_umi2(umi2, vec)
        self._encode_uei(uei, vec)

        # Pack the binary vector into bytes
        res = np.packbits(vec, bitorder="little")

        return res

    @overload
    def decode(
        self, bitvector: np.ndarray, skip_uei: Literal[True]
    ) -> tuple[bytes, bytes]: ...

    @overload
    def decode(
        self, bitvector: np.ndarray, skip_uei: Literal[False]
    ) -> tuple[bytes, bytes, bytes]: ...

    def decode(
        self, bitvector: npt.NDArray[np.uint8] | bytes, skip_uei: bool = False
    ) -> tuple[bytes, bytes] | tuple[bytes, bytes, bytes]:
        """Unpack the 84-bit UMI1 bitvector from a byte array.

        :param bitvector: the 256-bit vector to unpack
        :param skip_uei: whether to skip unpacking the UEI
        """
        if isinstance(bitvector, bytes):
            array_view = np.frombuffer(bitvector, dtype=np.uint8, count=len(bitvector))
        else:
            array_view = bitvector

        vec = np.unpackbits(array_view, bitorder="little")

        umi1_vec_slice = vec[self._umi1_slice].reshape((self._umi1_len, 3))
        packed_umi1 = np.packbits(umi1_vec_slice, axis=1, bitorder="little")
        umi1_array = np.take(_DECODING_LUT, packed_umi1, axis=0)
        umi1 = umi1_array.tobytes()

        umi2_vec_slice = vec[self._umi2_slice].reshape((self._umi2_len, 3))
        packed_umi2 = np.packbits(umi2_vec_slice, axis=1, bitorder="little")
        umi2_array = np.take(_DECODING_LUT, packed_umi2, axis=0)
        umi2 = umi2_array.tobytes()

        if skip_uei:
            return umi1, umi2

        uei_vec_slice = vec[self._uei_slice].reshape((self._uei_len, 3))
        packed_uei = np.packbits(uei_vec_slice, axis=1, bitorder="little")
        uei_array = np.take(_DECODING_LUT, packed_uei, axis=0)
        uei = uei_array.tobytes()

        return umi1, umi2, uei

    def encode_umi(self, umi: bytes):
        """Encode a single umi into a 3-bit per nucleotide encoded 128-bit vector.

        Each nucleotide is packed into 3 bits, so 8 nucleotides per 24-bit
        The output is padded to 128 bits since that is a common SIMD vector length
        and thus commonly used for efficient similarity search. (eg. in FAISS).

        :returns: a 16-byte/128-bits array of packed nucleotides
        :raises ValueError: if the UMI is longer than 32 nucleotides
        """
        if len(umi) > 40:
            raise ValueError("UMI cannot be longer than 40 nucleotides")

        vec = np.zeros(128, dtype=np.uint8)
        self._encode_umi1(umi, vec)

        # Pack the binary vector into bytes
        res = np.packbits(vec, bitorder="little")

        return res.tobytes()

    def decode_umi(self, umi_bytes: npt.NDArray[np.uint8] | bytes) -> bytes:
        """Decode a 3-bit encoded 128-bit umi vector into a nucleotide sequence.

        :param umi_bytes: the 128-bit umi vector
        :returns: the decoded nucleotide sequence
        :raises ValueError: if the input vector is not 128 bits long
        """
        if isinstance(umi_bytes, bytes):
            bytes_view = np.frombuffer(umi_bytes, dtype=np.uint8, count=len(umi_bytes))
        else:
            bytes_view = umi_bytes

        vec = np.unpackbits(bytes_view, bitorder="little")

        umi_vec_slice = vec[self._umi1_slice].reshape((self._umi1_len, 3))
        packed_umi = np.packbits(umi_vec_slice, axis=1, bitorder="little")
        umi_array = np.take(_DECODING_LUT, packed_umi, axis=0)
        umi = umi_array.tobytes()
        return umi

    def compress_umi_embedding(
        self, bitvector: npt.NDArray[np.uint8] | bytes
    ) -> np.uint64:
        """Compress the 3-bit per nucleotide into a more classical 2-bit embedding.

        The nucleotides encoded in this space are not equidistant anymore, but it is more compact.

        :param bitvector: the 128-bit umi vector
        :returns: the new 2-bit embedding for given input
        :raises ValueError: if the input vector is not 128 bits long
        """
        return self._compress_3bit_embedding(bitvector, 12, 28)

    def compress_uei_embedding(
        self, bitvector: npt.NDArray[np.uint8] | bytes
    ) -> np.uint64:
        """Compress the 3-bit per nucleotide embedded UEI into a 2-bit embedding."""
        return self._compress_3bit_embedding(bitvector, 8, 15)

    def _compress_3bit_embedding(
        self, bitvector: npt.NDArray[np.uint8] | bytes, n: int, decoded_len: int
    ) -> np.uint64:
        """Compress the 3-bit per nucleotide into a more classical 2-bit embedding.

        The nucleotides encoded in this space are not equidistant anymore, but it is more compact.

        Args:
            bitvector: A 3bit encoded byte array
            n: The length of encoded bytes without padding.
            decoded_len: The length of the uncompressed sequence

        Returns:
            the new 2-bit embedding for given input

        Raises:
            ValueError: if the input vector is not 128 bits long

        """
        array_view: npt.NDArray[np.uint8]

        if isinstance(bitvector, bytes):
            input_len = min(len(bitvector), n)
            array_view = np.zeros(n, dtype=np.uint8)
            array_view[:input_len] = np.frombuffer(
                bitvector, dtype=np.uint8, count=input_len
            )
        else:
            array_view = bitvector

        bitslice = slice(0, decoded_len * 3)

        vec = np.unpackbits(array_view, bitorder="little")
        vec_slice = vec[bitslice].reshape((decoded_len, 3))
        numeric_nucl_array = np.packbits(vec_slice, axis=1, bitorder="little")

        out = np.take(_RECODING_3BIT_TO_2BIT_LUT, numeric_nucl_array, axis=0)
        out = np.squeeze(out, 1)

        padded_2bit = np.zeros((8,), dtype=np.uint8)
        tmp = np.packbits(out, bitorder="little")
        padded_2bit[0 : len(tmp)] = tmp
        out_int = np.frombuffer(padded_2bit, dtype=np.uint64, count=1)[0]
        return out_int


class BarcodeDemuxer(abc.ABC):
    """An abstract base class for BarcodeDemuxer steps."""

    def __init__(self, assay: PNAAssay, panel: PNAAntibodyPanel):
        """Initialize a BarcodeDemuxer object.

        :param assay: the assay design
        :param panel: the antibody panel
        """
        self.assay = assay
        self.panel = panel
        self._pid1_slice = slice(*get_position_in_parent(assay, "pid-1"))
        self._pid2_slice = slice(*get_position_in_parent(assay, "pid-2"))
        self._umi1_slice = slice(*get_position_in_parent(assay, "umi-1"))
        self._umi2_slice = slice(*get_position_in_parent(assay, "umi-2"))
        self._uei_slice = slice(*get_position_in_parent(assay, "uei"))

        self._records_handled = 0
        self._records_written = 0
        self._output_groups_buffer: dict[int, DemuxRecordBatch] = defaultdict(
            DemuxRecordBatch
        )

        # 4 bits per nucleotide into bytes
        self._packed_umi1_len = (
            (self._umi1_slice.stop - self._umi1_slice.start) * 4 // 8
        )
        self._packed_umi2_len = (
            (self._umi2_slice.stop - self._umi2_slice.start) * 4 // 8
        )
        self._packed_uei_len = (self._uei_slice.stop - self._uei_slice.start) * 4 // 8

        self._embedding = PNAEmbedding(assay)

    def _serialize_group(
        self, group_id: int, group: DemuxRecordBatch
    ) -> tuple[int, bytes]:
        """Serialize and send a group to the writer process."""
        serialized_batch = group.serialize()
        self._records_written += len(group)
        group.clear()
        return (group_id, serialized_batch)

    def flush(self):
        """Return all remaining groups."""
        for group_id, group in self._output_groups_buffer.items():
            if len(group) > 0:
                yield self._serialize_group(group_id, group)
                group.clear()

        self._output_groups_buffer.clear()
        self._records_handled = 0
        self._records_written = 0

    @abc.abstractmethod
    def __call__(
        self, read: SequenceRecord
    ) -> list[tuple[int, bytes]] | tuple[int, bytes] | None:
        """Find the nearest antibody for a given barcode.

        :param read: the input read to process
        :returns: A tuple or list of tuples with a barcode_group_id and a serialized batch of records
            The serialized batch is encoded in the pyarrow RecordBatch ipc format.
        """
        raise NotImplementedError


class IndependentBarcodeDemuxer(BarcodeDemuxer):
    """Find the nearest antibody for a given barcode.

    The marker1 and marker2 barcode are grouped independently.
    Each read will be present twice in the output, once for each barcode.
    """

    def __init__(
        self,
        assay: PNAAssay,
        panel: PNAAntibodyPanel,
        marker1_groups: dict[str, int],
        marker2_groups: dict[str, int],
    ):
        """Initialize the BarcodeIdentifier object.

        :param assay: the assay design
        :param panel: the antibody panel
        :param marker1_groups: the mapping of marker1 to group id
        :param marker2_groups: the mapping of marker2 to group id
        """
        super().__init__(assay, panel)
        self.marker1_groups = marker1_groups
        self.marker2_groups = marker2_groups

    def __call__(self, read: SequenceRecord) -> list[tuple[int, bytes]] | None:
        """Find the nearest antibody for a given barcode.

        :param read: the read to process
        :return: the read with the antibody information added
        """
        if read.comment is None:
            raise BarcodeDemuxingError("No comment found in read")

        pid_comment = read.comment.split(" ")[-1]
        parts = pid_comment.split(":")
        if len(parts) != 2:
            raise BarcodeDemuxingError(f"Invalid PID comment: {pid_comment}")

        pid1, pid2 = pid_comment.split(":")
        # pid1, pid2 = self.record_group_mapping.get(read.id, (None, None))
        if pid1 is None or pid2 is None:
            return None

        read_bytes = read.sequence.encode("ascii")
        umi1 = read_bytes[self._umi1_slice]
        umi2 = read_bytes[self._umi2_slice]
        uei = read_bytes[self._uei_slice]

        vector = self._embedding.encode(umi1, umi2, uei)

        marker1_group_id = self.marker1_groups.get(pid1)
        marker2_group_id = self.marker2_groups.get(pid2)

        if marker1_group_id is None or marker2_group_id is None:
            raise BarcodeDemuxingError(
                f"One of the markers was not found in panel: {pid1}, {pid2}"
            )

        pid1_idx = self.panel.markers.index(pid1)
        pid2_idx = self.panel.markers.index(pid2)

        marker1_group = self._output_groups_buffer[marker1_group_id]
        marker1_group.add_record(pid1_idx, pid2_idx, vector)

        marker2_group = self._output_groups_buffer[marker2_group_id]
        marker2_group.add_record(pid1_idx, pid2_idx, vector)

        results = []
        if len(marker1_group) == marker1_group.capacity():
            results.append(self._serialize_group(marker1_group_id, marker1_group))

        if len(marker2_group) == marker2_group.capacity():
            results.append(self._serialize_group(marker2_group_id, marker2_group))

        return results or None


class PairedBarcodeDemuxer(BarcodeDemuxer):
    """Demux reads according to (PID1, PID2) groups.

    Each (PID1, PID2) pair is assigned to group id.

    """

    def __init__(
        self,
        assay: PNAAssay,
        panel: PNAAntibodyPanel,
        supergroups: dict[tuple[str, str], int],
    ):
        """Initialize the PairedBarcodeDemuxer object.

        :param assay: the assay design
        :param panel: the antibody panel
        :param supergroups: the mapping of (PID1, PID2) pairs to supergroup ids
        """
        super().__init__(assay, panel)
        self._supergroups = supergroups
        self._buffer_len = 10_000

    def __call__(self, read: SequenceRecord) -> tuple[int, bytes] | None:
        """Find the nearest antibody for a given barcode.

        :param read: the read to process
        :return: the read with the antibody information added
        """
        if read.comment is None:
            raise BarcodeDemuxingError("No comment found in read")

        pid_comment = read.comment.split(" ")[-1]
        parts = pid_comment.split(":")
        if len(parts) != 2:
            raise BarcodeDemuxingError(f"Invalid PID comment: {pid_comment}")

        pid1, pid2 = parts
        if pid1 is None or pid2 is None:
            return None

        read_bytes = read.sequence.encode("ascii")
        umi1 = read_bytes[self._umi1_slice]
        umi2 = read_bytes[self._umi2_slice]
        uei = read_bytes[self._uei_slice]
        vector = self._embedding.encode(umi1, umi2, uei)
        group_id = self._supergroups[(pid1, pid2)]

        if group_id is None:
            return None

        pid1_idx = self.panel.markers.index(pid1)
        pid2_idx = self.panel.markers.index(pid2)

        group = self._output_groups_buffer[group_id]
        group.add_record(pid1_idx, pid2_idx, vector)
        self._records_handled += 1

        if len(group) == group.capacity():
            return self._serialize_group(group_id, group)

        return None



"""Analysis engine capable of running a list of analysis functions on each component in a pixeldataset.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
from collections import defaultdict
from functools import partial
from queue import Queue
from typing import Callable, Iterable, Protocol

import pandas as pd

from pixelator.common.utils import (
    get_process_pool_executor,
)
from pixelator.mpx.graph import Graph
from pixelator.mpx.pixeldataset import PixelDataset

logger = logging.getLogger(__name__)


class PerComponentAnalysis(Protocol):
    """Protocol for analysis functions that are run on each component in a PixelDataset."""

    ANALYSIS_NAME: str

    def run_on_component(self, component: Graph, component_id: str) -> pd.DataFrame:
        """Run the analysis on this component."""
        ...

    def concatenate_data(self, data: Iterable[pd.DataFrame]) -> pd.DataFrame:
        """Concatenate the data. Override this if you need custom concatenation behavior."""
        try:
            scores = pd.concat(data, axis=0)
            return scores
        except ValueError as error:
            logger.error(f"No data was found to compute {self.ANALYSIS_NAME}")
            raise error

    def post_process_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Post process the data (e.g. adjust p-values). Override this if your data needs post processing."""
        return data

    def add_to_pixel_dataset(
        self, data: pd.DataFrame, pxl_dataset: PixelDataset
    ) -> PixelDataset:
        """Add the data in the right place in the pxl_dataset."""
        ...

    def parameters(self) -> dict:
        """Return the parameters of the `PerComponentAnalysis`.

        This is used e.g. to store the metadata the parameters of the analysis
        in the run metadata.
        """
        return {f"{self.ANALYSIS_NAME}": vars(self)}


class _AnalysisManager:
    """Analysis manager that can run a number of analysis across a stream of components.

    The analysis manager is responsible for hooking up the analysis functions and making
    them run on each component in the stream. The main workflow it uses is outlined in the
    `execute` method.
    """

    def __init__(
        self,
        analysis_to_run: Iterable[PerComponentAnalysis],
        component_stream: Iterable[tuple[str, Graph]],
    ):
        self.analysis_to_run = {
            analysis.ANALYSIS_NAME: analysis for analysis in analysis_to_run
        }
        self.component_stream = component_stream

    def _prepare_computation(
        self,
    ) -> Iterable[tuple[str, Callable[[Graph, str], pd.DataFrame]]]:
        for component_id, component_graph in self.component_stream:
            for _analysis_name, analysis in self.analysis_to_run.items():
                yield (
                    _analysis_name,
                    partial(
                        analysis.run_on_component,
                        component=component_graph,
                        component_id=component_id,
                    ),
                )

    def _execute_computations_in_parallel(self, prepared_computations):
        futures = Queue()
        with get_process_pool_executor() as executor:
            for analysis_name, func in prepared_computations:
                logger.debug("Putting %s in the queue for analysis", analysis_name)
                future = executor.submit(func)
                futures.put((analysis_name, future))

            while not futures.empty():
                key, future = futures.get()
                if future.done():
                    logger.debug("Future for %s is done", key)
                    yield (key, future.result())
                else:
                    futures.put((key, future))

    def _post_process(self, per_component_results):
        concatenated_data = defaultdict(list)
        for key, data in per_component_results:
            concatenated_data[key].append(data)

        for key, data_list in concatenated_data.items():
            yield (
                key,
                self.analysis_to_run[key].post_process_data(
                    self.analysis_to_run[key].concatenate_data(data_list)
                ),
            )

    def _add_to_pixel_dataset(self, post_processed_data, pxl_dataset: PixelDataset):
        for key, data in post_processed_data:
            pxl_dataset = self.analysis_to_run[key].add_to_pixel_dataset(
                data, pxl_dataset
            )
        return pxl_dataset

    def execute(self, pixel_dataset) -> PixelDataset:
        """Execute the analysis on the provided pixel dataset."""
        prepared_computations = self._prepare_computation()
        per_component_results = self._execute_computations_in_parallel(
            prepared_computations
        )
        post_processed_data = self._post_process(per_component_results)
        pxl_dataset_with_results = self._add_to_pixel_dataset(
            post_processed_data, pixel_dataset
        )
        return pxl_dataset_with_results


def edgelist_to_component_stream(
    dataset: PixelDataset, use_full_bipartite: bool
) -> Iterable[tuple[str, Graph]]:
    """Convert the edgelist in the dataset to a stream component ids and their component graphs."""
    for component_id, component_df in (
        dataset.edgelist_lazy.collect()
        .partition_by(by="component", as_dict=True)
        .items()
    ):
        yield (
            str(component_id[0]),  # component id is a tuple here, hence the [0]
            Graph.from_edgelist(
                edgelist=component_df.lazy(),
                add_marker_counts=True,
                simplify=True,
                use_full_bipartite=use_full_bipartite,
            ),
        )


def run_analysis(
    pxl_dataset: PixelDataset,
    analysis_to_run: list[PerComponentAnalysis],
    use_full_bipartite: bool = False,
) -> PixelDataset:
    """Run the provided list of `PerComponentAnalysis` on the components in the `pxl_dataset`.

    :param pxl_dataset: The PixelDataset to run the analysis on.
    :param analysis_to_run: A list of `PerComponentAnalysis` to run on the components in the `pxl_dataset`.
    :param use_full_bipartite: Whether to use the full bipartite graph when creating the components.
    :returns: A `PixelDataset` instance with the provided analysis added to it.
    """
    if not analysis_to_run:
        logger.warning("No analysis functions were provided")
        return pxl_dataset

    analysis_manager = _AnalysisManager(
        analysis_to_run,
        component_stream=edgelist_to_component_stream(
            pxl_dataset, use_full_bipartite=use_full_bipartite
        ),
    )
    pxl_dataset = analysis_manager.execute(pxl_dataset)
    return pxl_dataset



"""Module that contains functions for processing and QC any MPX amplicon design.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
import subprocess
from pathlib import Path
from subprocess import CalledProcessError

# List is used as type hint in comment
from typing import List  # noqa: F401

logger = logging.getLogger(__name__)


def qc_fastq(
    input: str,
    output: str,
    failed: str,
    report: str,
    metrics: str,
    design: str,
    n_limit: int,
    trim_front: int,
    trim_tail: int,
    min_length: int,
    max_length: int,
    threads: int,
    avg_qual: int,
    dedup: bool,
    remove_polyg: bool,
    verbose: bool,
) -> None:
    """Filter reads from fastq files on the provided quality criteria.

    This function is a wrapper around `fastp` to pre-process a `fastq`
    file. Duplicated sequences are removed, polyG sequences of length >= 5
    are trimmed. Some filters are used to discard reads (maximum and minimum
    read length and the maximum number of Ns in a read).

    :param input: the path to the fastq file
    :param output: the path to the output file (processed)
    :param failed: the path to the failed file (discarded reads)
    :param report: the path to the html report file
    :param metrics: the path to the json metrics file
    :param design: the design used in the config file
    :param n_limit: the number of Ns to use a cutoff to discard a read
    :param trim_front: the number of bases to trim in the front
    :param trim_tail: the number of bases to trim in the tail
    :param min_length: the minimum length for the reads
    :param max_length: the maximum length for the reads
    :param threads: the number of threads to use
    :param avg_qual: the minimum avg quality of a read
    :param dedup: remove duplicated reads when true
    :param remove_polyg: remove PolyG sequences (length 10 or more)
    :param verbose: run in verbose mode when true
    :returns: None
    :raises ValueError: raises an exception
            OSError: raises an exception
            CalledProcessError: raises an exception
            RuntimeError: raises an exception
    """
    args = [
        "fastp",
        "--disable_adapter_trimming",
        "--overrepresentation_analysis",
        "--thread",
        str(threads),
        "--n_base_limit",
        str(n_limit),
        "--trim_front1",
        str(trim_front),
        "--trim_tail1",
        str(trim_tail),
        "--max_len1",
        str(max_length),
        "--length_required",
        str(min_length),
        "--failed_out",
        failed,
        "--html",
        report,
        "--json",
        metrics,
        "--out1",
        output,
        "--average_qual",
        str(avg_qual),
        "--in1",
        input,
    ]  # type: List[str]

    if remove_polyg:
        args += ["--trim_poly_g", "--poly_g_min_len", "10"]
    if dedup:
        args += ["--dedup"]
    if verbose:
        args += ["--verbose"]

    logger.debug("Invoking fastp: %s", " ".join(args))

    try:
        proc = subprocess.Popen(  # type: ignore
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            shell=False,
        )
        (stdout, errmsg) = proc.communicate()
    except ValueError as exc:
        logger.error("ERROR running fastp. Incorrect arguments")
        raise exc
    except OSError as exc:
        logger.error("ERROR running fastp. Executable not found")
        raise exc
    except CalledProcessError as exc:
        logger.error("ERROR running fastp. Program returned error")
        raise exc

    error_string = errmsg.decode("utf-8")

    # check output file exists
    if not Path(output).is_file():
        error = (
            f"ERROR running fastp.\nOutput file not present {output}\n{error_string}\n"
        )
        logger.error(error)
        raise RuntimeError(error)

    # Propagate unexpected fastp errors
    if proc.returncode != 0:
        msg = f"Fastp encountered an error: \n\n {error_string} \n"
        logger.error(msg)
        raise RuntimeError(msg)


def adapter_qc_fastq(
    input: str,
    output: str,
    failed: str,
    report: str,
    mismatches: float,
    pbs1: str,
    pbs2: str,
    cores: int,
    verbose: bool,
) -> None:
    """Filter reads from fastq files that do not have PBS1/2 sequences.

    This function is a wrapper around `cutadapt` to process a `fastq`
    file with molecular pixelation data. The provided PBS1/2 sequences
    will be searched in the reads and only the reads that contain both
    sequences will be kept.

    :param input: the path to the fastq file (must contain PBS1/2 sequences)
    :param output: the path to the output file (processed)
    :param failed: the path to the failed file (discarded)
    :param report: the path to the json report
    :param mismatches: the number of mismatches allowed (0.1 - 0.9)
    :param pbs1: the PBS1 sequence
    :param pbs2: the PBS2 sequence
    :param cores: the number of threads to use
    :param verbose: run in verbose mode when true
    :returns: None
    :raises ValueError: raises an exception
            OSError: raises an exception
            CalledProcessError: raises an exception
            RuntimeError: raises an exception
    """
    min_overlap_pbs1 = len(pbs1) - int(mismatches * len(pbs1))
    min_overlap_pbs2 = len(pbs2) - int(mismatches * len(pbs2))
    args = [
        "cutadapt",
        "-e",
        str(mismatches),
        "--adapter",
        f"{pbs2};required;min_overlap={min_overlap_pbs2}...{pbs1};required;min_overlap={min_overlap_pbs1}",
        "--cores",
        str(cores),
        "--action=none",
        "--no-indels",
        "--untrimmed-output",
        failed,
        "--json",
        report,
        "--output",
        output,
        input,
    ]  # type: List[str]

    if verbose:
        args += ["--debug"]

    logger.debug("Invoking cutadapt: %s", " ".join(args))

    try:
        proc = subprocess.Popen(  # type: ignore
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            shell=False,
        )
        (stdout, errmsg) = proc.communicate()
    except ValueError as exc:
        logger.error("ERROR running cutadapt. Incorrect arguments")
        raise exc
    except OSError as exc:
        logger.error("ERROR running cutadapt. Executable not found")
        raise exc
    except CalledProcessError as exc:
        logger.error("ERROR running cutadapt. Program returned error")
        raise exc

    error_string = errmsg.decode("utf-8")

    # check output file exists

    if not Path(output).is_file():
        error = (
            f"ERROR running cutadapt.\nOutput file not present "
            f"{output}\n{stdout.decode('utf-8')}\n{errmsg.decode('utf-8')}"
        )
        logger.error(error)
        raise RuntimeError(error)

    # Propagate unexpected cutadapt errors
    if proc.returncode != 0:
        msg = f"Cutadapt encountered an error: \n\n {error_string} \n"
        logger.error(msg)
        raise RuntimeError(msg)



"""Collapse molecules from a demultiplexed PNA dataset.

Copyright © 2024 Pixelgen Technologies AB.
"""

import dataclasses
import logging
import math
import multiprocessing
import time
import typing
from contextlib import contextmanager
from functools import cached_property
from multiprocessing.managers import SharedMemoryManager
from multiprocessing.shared_memory import SharedMemory
from pathlib import Path
from typing import Generator

import joblib
import numpy as np
import numpy.typing as npt
import polars as pl
import pyarrow
import pyarrow as pa
import pyarrow.compute
import pyarrow.parquet

from pixelator.common.report.models import SummaryStatistics
from pixelator.pna.collapse.adjacency import (
    build_network_cluster,
    build_network_directional,
)
from pixelator.pna.collapse.paired.statistics import (
    CollapseStatistics,
    MarkerLinkGroupStats,
)
from pixelator.pna.collapse.utilities import (
    _find_connected_components_cluster,
    _find_connected_components_directional,
    build_binary_index,
)
from pixelator.pna.config import PNAAntibodyPanel, PNAAssay
from pixelator.pna.demux.barcode_demuxer import PNAEmbedding
from pixelator.pna.utils import pack_2bits
from pixelator.pna.utils.two_bit_encoding import pack_4bits

logger = logging.getLogger("collapse")


def _collect_label_array_indices(
    labels: npt.NDArray[np.int32], n_components: int
) -> npt.NDArray[np.object_]:
    """Collect an array with a label for each item into an array of indices for each label.

    Args:
        labels: The labels for each item.
        n_components: Number of labels in the label array.

    Returns:
        An array of indices for each label.

    """
    # Single pass over the labels to collect the indices of each connected component
    groups = np.ndarray(shape=(n_components,), dtype=object)  # type: ignore

    for idx, value in enumerate(labels):
        g = groups[value]
        if g is None:
            groups[value] = [idx]
        else:
            g.append(idx)

    return groups


def _split_chunks(
    n_components: int, chunk_size: int
) -> Generator[tuple[int, int], None, None]:
    """Split a range [0, n_components) into chunks of size chunk_size.

    This is a generator function that yields the start and stop indices of each chunk.
    The last chunk may be smaller than chunk_size.

    Args:
        n_components: The size of the input range.
        chunk_size: The size of each chunk.

    Yields:
        A tuple of (start, stop) indices for each chunk.

    """
    pos = 0
    complete_chunks = n_components // chunk_size

    for i in range(complete_chunks):
        yield (pos, pos + chunk_size)
        pos += chunk_size

    yield (pos, n_components)


@dataclasses.dataclass(slots=True, frozen=True)
class _DistributedResults:
    start: int
    stop: int

    umi1: npt.ArrayLike
    umi2: npt.ArrayLike
    uei: npt.ArrayLike

    read_count: npt.ArrayLike
    corrected_reads: int


MoleculeCollapserAlgorithm = typing.Literal["directional", "cluster"]


class MoleculeCollapser:
    """Collapse molecules from a demultiplexed PNA dataset.

    This class is responsible for collapsing molecules from a demultiplexed PNA dataset.
    Collapsing is parallelized for each marker pair so parallelizing collapsing over multiple
    independent groups is not needed.
    """

    _collapsed_schema = pa.schema(
        [
            pa.field("umi1", pa.uint64()),
            pa.field("umi2", pa.uint64()),
            pa.field("uei", pa.uint64()),
            pa.field("read_count", pa.int64()),
        ]
    )

    _output_schema = pa.schema(
        [
            pa.field("marker_1", pa.large_string()),
            pa.field("marker_2", pa.large_string()),
            pa.field("umi1", pa.uint64()),
            pa.field("umi2", pa.uint64()),
            pa.field("read_count", pa.int64()),
            pa.field("uei_count", pa.int64()),
        ]
    )

    def __init__(
        self,
        assay: PNAAssay,
        panel: PNAAntibodyPanel,
        output: Path,
        max_mismatches: int | float = 0.1,
        algorithm: MoleculeCollapserAlgorithm = "directional",
        threads: int = -1,
        logger: logging.Logger | None = None,
        min_parallel_chunk_size: int = 2_000,
    ):
        """Initialize a MoleculeCollapser.

        Args:
            assay: The assay configuration.
            panel: The antibody panel configuration.
            output: The output path for the collapsed molecules.
            max_mismatches: The maximum number of mismatches allowed when collapsing molecules.
                Either an integer >= 1 or a float in the range [0, 1).
            algorithm: The algorithm to use for collapsing molecules. Either "directional" or "cluster".
            threads: The number of threads to use for parallel processing.
            logger: The logger to use for output. The default is a logger named "collapse".
            min_parallel_chunk_size: The minimum number of connected components to process in parallel.
                Components below this size will be processed serially.

        """
        self.assay = assay
        self.panel = panel
        self.algorithm = algorithm

        if max_mismatches >= 1:
            if isinstance(max_mismatches, float):
                raise ValueError(
                    "max_mismatches must be either an integer value > 1"
                    " or a float value between 0 and 1."
                )
            self.max_mismatches = max_mismatches
        else:
            molecule_len = (
                self.assay.get_region_by_id("umi-1").max_len
                + self.assay.get_region_by_id("umi-2").max_len
                + self.assay.get_region_by_id("uei").max_len
            )
            self.max_mismatches = math.ceil(molecule_len * max_mismatches)

        self.threads = threads
        self._memory_manager = SharedMemoryManager()
        self._parallel_worker = joblib.Parallel(n_jobs=threads, return_as="list")
        self._logger = logger or logging.getLogger("collapse")

        self._writer = pyarrow.parquet.ParquetWriter(
            output, schema=self._output_schema, compression="zstd", compression_level=6
        )
        self._marker_group_state = {
            "db_shm": None,
            "read_counts_shm": None,
        }

        self._bitvector = PNAEmbedding(self.assay)
        self._stats = CollapseStatistics()

        self.min_parallel_chunk_size = min_parallel_chunk_size

    def __enter__(self):
        """Enter the context manager for the MoleculeCollapser.

        This wraps the context managers for the shared memory manager, parallel worker pool,
        and the parquet writer.
        """
        self._memory_manager = self._memory_manager.__enter__()
        self._parallel_worker = self._parallel_worker.__enter__()
        self._writer = self._writer.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager for the MoleculeCollapser.

        This wraps the context managers for the shared memory manager, parallel worker pool,
        and the parquet writer.
        """
        self._memory_manager.__exit__(exc_type, exc_val, exc_tb)
        self._parallel_worker.__exit__(exc_type, exc_val, exc_tb)
        self._writer.__exit__(exc_type, exc_val, exc_tb)

    @cached_property
    def _effective_threads(self):
        if self.threads <= -1:
            return multiprocessing.cpu_count() + 1 - abs(self.threads)

        return self.threads

    @property
    def _current_db_memory(self) -> SharedMemory | None:
        return self._marker_group_state["db_shm"]

    @_current_db_memory.setter
    def _current_db_memory(self, buffer):
        self._marker_group_state["db_shm"] = buffer

    @property
    def _current_read_counts_memory(self) -> SharedMemory | None:
        return self._marker_group_state["read_counts_shm"]

    @_current_read_counts_memory.setter
    def _current_read_counts_memory(self, buffer):
        self._marker_group_state["read_counts_shm"] = buffer

    @contextmanager
    def _init_binary_vectors(self, data):
        """Return a contextmanager for the shared memory containing embeddings.

        Args:
            data: The data to initialize the binary vectors with.

        Yields:
            The binary vectors.

        """
        vector_length = 32
        shm_buffer = self._memory_manager.SharedMemory(size=(len(data) * vector_length))
        self._current_db_memory = shm_buffer

        db = np.ndarray(
            (len(data), vector_length), dtype=np.uint8, buffer=shm_buffer.buf
        )
        db[:, :] = 0

        for idx, vector in enumerate(data["molecule"]):
            db[idx, :] = np.frombuffer(vector, dtype=np.uint8, count=vector_length)

        yield db

        self._current_db_memory.unlink()
        self._current_db_memory = None

    @contextmanager
    def _init_read_counts(self, data):
        """Return a contextmanager for the shared memory containing read counts.

        Args:
            data: The data to initialize the read counts with.

        Yields:
            The read counts.

        """
        n_molecules = len(data)
        shm_buffer = self._memory_manager.SharedMemory(
            size=(n_molecules * np.dtype(np.uint64).itemsize)
        )
        read_counts = np.ndarray(n_molecules, dtype=np.uint64, buffer=shm_buffer.buf)
        read_counts[:] = data["read_count"].to_numpy()
        self._current_read_counts_memory = shm_buffer

        yield read_counts

        self._current_read_counts_memory.unlink()
        self._current_read_counts_memory = None

    @cached_property
    def max_hamming_mismatches(self) -> int:
        """Return the maximum hamming distance for the number of allowed mismatches."""
        return self.max_mismatches * 2

    @staticmethod
    def _get_db_from_shared_memory(db_shm: SharedMemory, count: int):
        """Return a numpy array backed by a shared memory buffer for the embeddings.

        Args:
            db_shm: the shared memory buffer.
            count: the number of vectors in the buffer

        Returns:
            A numpy array containing the embeddings.

        """
        vector_len = 32
        db = np.frombuffer(db_shm.buf, dtype=np.uint8, count=count * vector_len)
        db.shape = (count, vector_len)
        return db

    @staticmethod
    def _get_counts_from_shared_memory(counts_shm: SharedMemory, count: int):
        """Return a numpy array backed by a shared memory buffer for the read counts.

        Args:
            counts_shm: The shared memory buffer.
            count: The number of read counts in the buffer.

        Returns:
            The numpy array.

        """
        counts = np.frombuffer(counts_shm.buf, dtype=np.int64, count=count)
        return counts

    def _record_group_serial_fn(
        self,
        output_idx,
        component_indices,
        db_shm,
        read_counts_shm,
        db_size,
        outputs,
        n_components,
    ):
        """Record a group of connected components serially.

        Args:
            output_idx: The output index.
            component_indices: The indices of the components.
            db_shm: The shared memory buffer for the database.
            read_counts_shm: The shared memory buffer for the read counts.
            db_size: The size of the database.
            outputs: The outputs.
            n_components: The number of components.

        Returns:
            The number of corrected reads.

        """
        db = self._get_db_from_shared_memory(db_shm, db_size)
        read_count = self._get_counts_from_shared_memory(read_counts_shm, db_size)

        # Aggregate the read_count for this connected component
        comp_read_counts = read_count[component_indices]
        total_read_count = np.sum(comp_read_counts)

        # Select the molecule with the most support as "representative" of the component
        representative_idx = np.argmax(comp_read_counts)
        comp_idx = component_indices[representative_idx]
        umi1, umi2, uei = self._bitvector.decode(db[comp_idx], skip_uei=False)

        umi1 = pack_2bits(umi1)
        umi2 = pack_2bits(umi2)
        uei = pack_4bits(uei)
        read_count = total_read_count
        corrected_reads = int(total_read_count - comp_read_counts[representative_idx])

        outputs[0][output_idx] = umi1
        outputs[1][output_idx] = umi2
        outputs[2][output_idx] = uei
        outputs[3][output_idx] = read_count

        return int(corrected_reads)

    @staticmethod
    def _record_group_worker_fn(
        subrange: tuple[int, int],
        component_indices,
        db_shm,
        read_counts_shm,
        db_size,
        embedding: PNAEmbedding,
    ):
        """Process a batch of connected components.

        The database and read counts are loaded from shared memory to reduce
        python multiprocessing IPC overhead.

        Args:
            subrange: The range of connected components to process.
                A tuple with the start and stop indices.
            component_indices: A list of lists containing the indices
                in the database and read counts vector for each connected component.
            db_shm: The shared memory buffer containing the binary vectors.
            read_counts_shm: The shared memory buffer containing the read counts.
            db_size: The size of the binary vectors memory buffer in bytes.
            embedding: The PNAEmbedding instance for encoding/decoding vectors.

        Returns:
            The distributed results.

        """
        cls = MoleculeCollapser

        db = cls._get_db_from_shared_memory(db_shm, db_size)
        read_count = cls._get_counts_from_shared_memory(read_counts_shm, db_size)
        corrected_reads = 0

        subrange_len = subrange[1] - subrange[0]
        umi1_out = np.zeros((subrange_len,), dtype=np.uint64)
        umi2_out = np.zeros((subrange_len,), dtype=np.uint64)
        uei_out = np.zeros((subrange_len,), dtype=np.uint64)
        reads_out = np.zeros((subrange_len,), dtype=np.int64)

        for idx in range(subrange_len):
            this_cmp_indices = component_indices[idx]

            # Aggregate the read_count for this connected component
            molecule_read_counts = read_count[this_cmp_indices]
            component_reads = np.sum(molecule_read_counts)

            # Select the molecule with the most support as "representative" of the component
            # TODO: Voting on a representative molecule per nucleotide?
            representative_idx = np.argmax(molecule_read_counts)
            comp_idx = this_cmp_indices[representative_idx]
            umi1_bytes, umi2_bytes, uei_bytes = embedding.decode(
                db[comp_idx], skip_uei=False
            )

            umi1 = pack_2bits(umi1_bytes)
            umi2 = pack_2bits(umi2_bytes)
            uei = pack_4bits(uei_bytes)
            corrected_reads += (
                component_reads - molecule_read_counts[representative_idx]
            )

            umi1_out[idx] = umi1
            umi2_out[idx] = umi2
            uei_out[idx] = uei
            reads_out[idx] = component_reads

        res = _DistributedResults(
            start=subrange[0],
            stop=subrange[1],
            umi1=umi1_out,
            umi2=umi2_out,
            uei=uei_out,
            read_count=reads_out,
            corrected_reads=int(corrected_reads),
        )
        return res

    def _process_molecule_graph(
        self, csgraph, local_stats
    ) -> tuple[pa.Table, MarkerLinkGroupStats]:
        """Determine connected components and collapse the molecules in each component.

        Args:
            csgraph: The sparse adjacency matrix of the connected components.
            local_stats: The statistics object for this marker pair.

        Returns:
            A tuple with a pyarrow Table with the collapsed molecules
            and the updated statistics object.

        """
        _logger = self._logger

        n_components: int
        labels: npt.NDArray[np.int32]

        if self.algorithm == "cluster":
            n_components, labels = _find_connected_components_cluster(csgraph)
        elif self.algorithm == "directional":
            n_components, labels = _find_connected_components_directional(csgraph)
        else:
            raise ValueError(f"Unknown algorithm: {self.algorithm}")

        n_molecules = csgraph.shape[0]

        # Single pass over the labels to collect the indices of each connected component
        groups = _collect_label_array_indices(labels, n_components)
        local_stats.cluster_size_distribution = np.bincount(
            np.vectorize(len)(groups)
        ).tolist()

        db_mem = self._current_db_memory
        rc_mem = self._current_read_counts_memory
        bitvector = self._bitvector

        def work_fn_parallel(subrange):
            start, stop = subrange
            res = MoleculeCollapser._record_group_worker_fn(
                subrange, groups[start:stop], db_mem, rc_mem, n_molecules, bitvector
            )
            return res

        # Allocate output vectors in shared memory or as normal numpy arrays
        if (
            self._effective_threads > 1
            and n_components >= 2 * self.min_parallel_chunk_size
        ):
            chunk_size = max(
                self.min_parallel_chunk_size, n_components // self._effective_threads
            )
            _logger.info(
                "Detecting unique molecules from %s connected components (parallel: [chunk_size=%s])",
                n_components,
                chunk_size,
            )
            subranges = list(_split_chunks(n_components, chunk_size=chunk_size))

            job_results: list[_DistributedResults] = self._parallel_worker(
                joblib.delayed(work_fn_parallel)(r) for r in subranges
            )
        else:
            _logger.info(
                "Determining collapsed read for %s connected components (serial)",
                n_components,
            )
            job_results = [work_fn_parallel((0, n_components))]

        umi1_arrays = []
        umi2_arrays = []
        uei_arrays = []
        read_count_arrays = []

        for r in job_results:
            umi1_arrays.append(r.umi1)
            umi2_arrays.append(r.umi2)
            uei_arrays.append(r.uei)
            read_count_arrays.append(r.read_count)

            local_stats.corrected_reads_count += r.corrected_reads

        table = pyarrow.Table.from_arrays(
            arrays=[
                pa.chunked_array(umi1_arrays),
                pa.chunked_array(umi2_arrays),
                pa.chunked_array(uei_arrays),
                pa.chunked_array(read_count_arrays),
            ],
            schema=self._collapsed_schema,
        )

        local_stats.collapsed_molecules_count = len(table)
        read_count_collapsed = table.column("read_count")
        local_stats.read_count_per_collapsed_molecule_stats = (
            SummaryStatistics.from_series(read_count_collapsed)
        )

        grouped = table.group_by(["umi1", "umi2"]).aggregate(
            [("read_count", "sum"), ("uei", "count_distinct")]
        )

        grouped = grouped.rename_columns(["umi1", "umi2", "read_count", "uei_count"])

        local_stats.unique_marker_links_count = len(grouped)

        read_count_grouped = grouped.column("read_count")
        local_stats.read_count_per_unique_marker_link_stats = (
            SummaryStatistics.from_series(read_count_grouped)
        )
        local_stats.uei_count_per_unique_marker_link_stats = (
            SummaryStatistics.from_series(grouped["uei_count"])
        )
        return grouped, local_stats

    def _process_marker_group(self, idx, num_groups, markers, data):
        """Process a group of markers.

        Args:
            idx: The index of the group.
            num_groups: The total number of groups.
            markers: The markers in the group.
            data: The data for the group.

        """
        starttime = time.time()
        _logger = self._logger

        marker1_name = self.panel.markers[markers[0]]
        marker2_name = self.panel.markers[markers[1]]

        input_molecules_count = len(data)
        input_reads_count = data["read_count"].sum()

        local_stats = MarkerLinkGroupStats(
            marker_1=marker1_name,
            marker_2=marker2_name,
            input_molecules_count=input_molecules_count,
            input_reads_count=input_reads_count,
        )

        _logger.info(
            "Processing %s molecules from markers [%s, %s] (%s/%s)",
            input_molecules_count,
            marker1_name,
            marker2_name,
            idx + 1,
            num_groups,
        )

        # Allocate shared memory for the binary vectors that make up the search space
        # and the read counts for each molecule. These are loaded with a context manager
        # to ensure proper cleanup
        with (
            self._init_binary_vectors(data) as db,
            self._init_read_counts(data) as read_counts,
        ):
            _logger.info(
                "Building binary index for [%s, %s] from %s molecules",
                marker1_name,
                marker2_name,
                len(db),
            )
            index = build_binary_index(db)

            _logger.info(
                "Querying binary index for [%s, %s] for similar molecules",
                marker1_name,
                marker2_name,
            )
            distances, indices = index.search(db, 48)

            _logger.info(
                "Building sparse adjacency matrix for [%s, %s]",
                marker1_name,
                marker2_name,
            )

            if self.algorithm == "cluster":
                adjacency = build_network_cluster(
                    distances, indices, read_counts, self.max_hamming_mismatches
                )
            elif self.algorithm == "directional":
                adjacency = build_network_directional(
                    distances, indices, read_counts, self.max_hamming_mismatches
                )
            else:
                raise ValueError(f"Unknown algorithm: {self.algorithm}")

            _logger.info(
                f"Collapsing clustered reads ({self.algorithm}) for [%s, %s]",
                marker1_name,
                marker2_name,
            )

            records, cluster_stats = self._process_molecule_graph(
                adjacency, local_stats
            )

            marker1_array = np.zeros(shape=(len(records)), dtype="<U16")
            marker2_array = np.zeros(shape=(len(records)), dtype="<U16")
            marker1_array[:] = marker1_name
            marker2_array[:] = marker2_name

            records = records.add_column(
                0,
                self._output_schema.field("marker_1"),
                pa.array(marker1_array, type=pa.large_string()),
            )
            records = records.add_column(
                1,
                self._output_schema.field("marker_2"),
                pa.array(marker2_array, type=pa.large_string()),
            )

            _logger.info(
                "Collapsed %s molecules into %s unique antibody links, correcting %s reads of %s total reads",
                input_molecules_count,
                len(records),
                cluster_stats.corrected_reads_count,
                cluster_stats.input_reads_count,
            )

            _logger.info("Streaming %s records to parquet", len(records))
            self._writer.write_table(records)

            _logger.info(
                "Completed processing %s molecules from markers [%s, %s] (%s/%s)",
                input_molecules_count,
                marker1_name,
                marker2_name,
                idx + 1,
                num_groups,
            )

        elapsed_time = time.time() - starttime
        self._stats.add_marker_stats(
            marker1_name,
            marker2_name,
            cluster_stats=cluster_stats,
            elapsed_time=elapsed_time,
        )

        return

    def process_file(self, path: Path) -> None:
        """Collapse molecules from a single parquet file containing molecular data.

        Args:
            path: The path to the parquet file.

        """
        _logger = self._logger

        _logger.info("Loading data from parquet: %s", str(path))
        df = pl.read_parquet(path)

        # Handle legacy data
        if "marker1" in df.columns or "marker2" in df.columns:
            df = df.rename({"marker1": "marker_1", "marker2": "marker_2"}, strict=False)

        self._stats.add_input_file(path, df.shape[0])

        _logger.info("Partitioning dataset by marker pairs")
        # Partition the data by the marker1 and marker2 columns and store each partition as a separate DataFrame
        partitions = df.partition_by(
            ["marker_1", "marker_2"], as_dict=True, include_key=False
        )

        num_partitions = len(partitions)
        _logger.info("Dataset partitioned into %s subsets", num_partitions)

        for idx, (markers, parts) in enumerate(partitions.items()):
            self._process_marker_group(idx, num_partitions, markers, parts)

    def statistics(self) -> CollapseStatistics:
        """Return the statistics collected during the collapse process.

        Returns:
            The collapse statistics.

        """
        return self._stats



"""Model for a collection of summary statistics of a distribution.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

import numpy as np
import pydantic

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


if TYPE_CHECKING:
    import pandas as pd
    import polars as pl


T = TypeVar("T")


class SummaryStatistics(pydantic.BaseModel):
    """A collection of summary statistics for a distribution."""

    mean: float
    std: float
    min: float
    q1: float
    q2: float
    q3: float
    max: float
    count: int

    @pydantic.computed_field(
        return_type=float,
        description="The interquartile range.",
    )
    def iqr(self) -> float:
        """Return the interquartile range."""
        return self.q3 - self.q1

    @classmethod
    def from_series(cls, distribution: pd.Series | pl.Series) -> Self:
        """Create a SummaryStatistics from a pandas or polars Series.

        :param distribution: The series to summarize.
        :return: A SummaryStatistics object from given series.
        """
        distribution = distribution.to_numpy()

        mean = np.mean(distribution)
        std = np.std(distribution)
        min = np.min(distribution)
        max = np.max(distribution)
        q1, q2, q3 = np.quantile(distribution, [0.25, 0.5, 0.75])
        count = len(distribution)

        return cls(
            mean=float(mean),
            std=float(std),
            q1=float(q1),
            q2=float(q2),
            q3=float(q3),
            max=float(max),
            min=float(min),
            count=int(count),
        )



"""Module for defining the HasCustomStatistics interface.

Copyright © 2024 Pixelgen Technologies AB.
"""

import abc
from typing import Any


class HasCustomStatistics(abc.ABC):
    """Base class enabling reducing arbitrary statistics from different workers.

    Pipeline steps that generate statistics should inherit from this class.
    """

    @abc.abstractmethod
    def get_statistics_name(self) -> str:
        """Return a string reprenting statistics generated by this Pipeline step."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_statistics(self) -> Any:
        """Return the statistics collected by this Pipeline step."""
        raise NotImplementedError



"""Main console script for pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

import atexit
import multiprocessing
import os
import sys
from pathlib import Path

import click
import yappi

from pixelator import __version__
from pixelator.common.utils import click_echo
from pixelator.mpx.cli.adapterqc import adapterqc
from pixelator.mpx.cli.amplicon import amplicon
from pixelator.mpx.cli.analysis import analysis
from pixelator.mpx.cli.annotate import annotate
from pixelator.mpx.cli.collapse import collapse
from pixelator.mpx.cli.common import AliasedOrderedGroup, logger
from pixelator.mpx.cli.demux import demux
from pixelator.mpx.cli.graph import graph
from pixelator.mpx.cli.layout import layout
from pixelator.mpx.cli.misc import list_single_cell_designs, list_single_cell_panels
from pixelator.mpx.cli.plugin import add_cli_plugins
from pixelator.mpx.cli.preqc import preqc
from pixelator.mpx.cli.report import report
from pixelator.mpx.logging import LoggingSetup


@click.group(cls=AliasedOrderedGroup, name="pixelator")
@click.version_option(__version__)
@click.option(
    "--verbose",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help="Show extended messages during execution",
)
@click.option(
    "--profile",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help="Activate profiling mode",
)
@click.option(
    "--log-file",
    required=False,
    default=None,
    type=click.Path(exists=False),
    help="The path to the log file (it is created if it does not exist)",
)
@click.option(
    "--cores",
    default=max(1, multiprocessing.cpu_count() - 1),
    required=False,
    type=click.INT,
    show_default=True,
    help="The number of cpu cores to use for parallel processing",
)
@click.pass_context
def main_cli(ctx, verbose: bool, profile: bool, log_file: str, cores: int):
    """Run the main CLI entrypoint for pixelator."""
    # early out if run in help mode
    if any(x in sys.argv for x in ["--help", "--version"]):
        return 0

    if verbose:
        logger.info("Running in VERBOSE mode")

    # activate profiling mode
    if profile:
        logger.info("Running in profiling mode")
        yappi.start()

        def exit():
            yappi.stop()
            logger.info("Profiling completed")
            processes = yappi.get_thread_stats()
            # Make sure to get profile metrics for each thread
            for p in processes:
                click_echo(f"Function stats for {p.name} {p.id}")
                yappi.get_func_stats(ctx_id=p.id).print_all()

        atexit.register(exit)

    # Pass arguments to other commands
    ctx.ensure_object(dict)

    # This registers the logger with it's context manager,
    # so that it is clean-up properly when the command is done.
    ctx.obj["LOGGER"] = ctx.with_resource(LoggingSetup(log_file, verbose=verbose))
    ctx.obj["VERBOSE"] = verbose
    ctx.obj["CORES"] = max(1, cores)

    # Read in environment variables
    duckdb_temp_dir = os.environ.get("PIXELATOR_DUCKDB_TEMP_DIR")
    if duckdb_temp_dir:
        ctx.obj["DUCKDB_TEMP_DIR"] = Path(duckdb_temp_dir)

    duckdb_tmp_dir_size = os.environ.get("PIXELATOR_DUCKDB_MAX_TEMP_DIR_SIZE")
    if duckdb_tmp_dir_size:
        ctx.obj["DUCKDB_MAX_TEMP_DIR_SIZE"] = duckdb_tmp_dir_size.strip()

    return 0


@main_cli.group(name="single-cell-mpx")
@click.option(
    "--list-designs",
    is_flag=True,
    metavar="",
    is_eager=True,
    expose_value=False,
    required=False,
    callback=list_single_cell_designs,
    help="List available designs and exit.",
)
@click.option(
    "--list-panels",
    is_flag=True,
    metavar="",
    is_eager=True,
    expose_value=False,
    required=False,
    callback=list_single_cell_panels,
    help="List available panels and exit.",
)
def single_cell_mpx():
    """Commands related to the molecular pixelator assay."""


# Add single-cell top level command to cli
main_cli.add_command(single_cell_mpx)

# Add single-cell commands
single_cell_mpx.add_command(amplicon)
single_cell_mpx.add_command(preqc)
single_cell_mpx.add_command(adapterqc)
single_cell_mpx.add_command(demux)
single_cell_mpx.add_command(collapse)
single_cell_mpx.add_command(graph)
single_cell_mpx.add_command(annotate)
single_cell_mpx.add_command(layout)
single_cell_mpx.add_command(analysis)
single_cell_mpx.add_command(report)


# Note that the import order here is intentional.
# The PNA commands needs to be imported here, rather than
# at the top, because otherwise things like importing
# the configs will not work as expected.
from pixelator.pna.cli.amplicon import amplicon
from pixelator.pna.cli.analysis import analysis
from pixelator.pna.cli.collapse import collapse
from pixelator.pna.cli.combine_collapse import combine_collapse
from pixelator.pna.cli.demux import demux
from pixelator.pna.cli.denoise import denoise
from pixelator.pna.cli.graph import graph
from pixelator.pna.cli.layout import layout
from pixelator.pna.cli.misc import (
    list_single_cell_pna_designs,
    list_single_cell_pna_panels,
)
from pixelator.pna.cli.report import report


@click.group()
@click.option(
    "--list-designs",
    is_flag=True,
    metavar="",
    is_eager=True,
    expose_value=False,
    required=False,
    callback=list_single_cell_pna_designs,
    help="List available designs and exit.",
)
@click.option(
    "--list-panels",
    is_flag=True,
    metavar="",
    is_eager=True,
    expose_value=False,
    required=False,
    callback=list_single_cell_pna_panels,
    help="List available panels and exit.",
)
def single_cell_pna():
    """Commands related to the network proximity assay."""
    pass


main_cli.add_command(single_cell_pna)

# Add all commands to the group
single_cell_pna.add_command(amplicon)
single_cell_pna.add_command(demux)
single_cell_pna.add_command(collapse)
single_cell_pna.add_command(graph)
single_cell_pna.add_command(denoise)
single_cell_pna.add_command(analysis)
single_cell_pna.add_command(layout)
single_cell_pna.add_command(report)
single_cell_pna.add_command(combine_collapse, name="combine-collapse")

# Add cli plugins as commands on top level
add_cli_plugins(main_cli)

if __name__ == "__main__":
    sys.exit(main_cli())  # pragma: no cover



"""Model for report data returned by the single-cell layout stage.

Copyright © 2024 Pixelgen Technologies AB.
"""

from __future__ import annotations

from pixelator.pna.report.models.base import SampleReport


class LayoutSampleReport(SampleReport):
    """Model for report data returned by the single-cell layout stage."""

    pass



"""Copyright © 2023 Pixelgen Technologies AB."""

import logging
from pathlib import Path

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseCollapseTestsMixin(BaseWorkflowTestMixin):
    """Base class for collapse command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_collapse_run"])
    ```
    """

    __stage_key__ = "collapse"

    @pytest.mark.dependency(name="test_collapse_run", depends=["test_demux_run"])
    def test_collapse_run(self):
        input_files = [
            str(f) for f in (self.workdir / "demux").glob("*processed*fastq.gz")
        ]
        panel = self.__get_data("panel")
        panel_file = self.__get_data("panel_file")
        design = self.__get_data("design")
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")

        command = [
            "--log-file",
            "collapse-pixelator.log",
            "single-cell-mpx",
            "collapse",
            "--output",
            ".",
            "--design",
            design,
            "--panel",
            str(Path(panel_file).name) if panel_file else panel,
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += input_files

        file_deps = list(input_files)
        if panel_file is not None:
            file_deps.append(panel_file)

        self.context.run_command("collapse", command, file_deps)
        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_collapse_run"])
    def test_collapse_logfile_exist(self):
        assert (self.workdir / "collapse-pixelator.log").is_file()

    @pytest.mark.dependency(depends=["test_collapse_run"])
    def test_collapse_results_folder_exists(self):
        assert (self.workdir / "collapse").is_dir()



"""Module with functions for created permuted data.

Copyright © 2024 Pixelgen Technologies AB.
"""

from time import time
from typing import Generator, Optional

import pandas as pd
import polars as pl
from numpy.random import Generator as RandomNumberGenerator
from numpy.random import default_rng


def _get_random_number_generator(
    random_seed: Optional[int] = None,
) -> RandomNumberGenerator:
    if not random_seed:
        random_seed = int(time())
    return default_rng(seed=random_seed)


def edgelist_permutations(
    edgelist_df: pl.DataFrame, n: int = 50, random_seed: Optional[int] = None
) -> Generator[pl.DataFrame, None, None]:
    """Generate `n` permutations of the markers in `edgelist_df`.

    :param edgelist_df: dataframe to use as basis of permutations
    :param n: number of permutations to generate, defaults to 50
    :param random_seed: set a seed to the random number generator
                        needed to make results deterministic, defaults to None
    :return: a generator that yields `n` permutations of the input dataframe
    :rtype: Generator[pl.DataFrame, None, None]
    """
    random_number_generator = _get_random_number_generator(random_seed)
    for _ in range(n):
        yield permute_edgelist(edgelist_df, random_number_generator)


def permute_edgelist(
    edgelist: pl.DataFrame,
    random_number_generator: Optional[RandomNumberGenerator] = None,
):
    """Permute markers in an edgelist.

    This function permutes the edgelist by shuffling the corresponding
    markers to umi1 and umi2 columns.

    :param edgelist: A DataFrame representing the edgelist
    :param n_permutations: The number of permutations to perform
    :returns: A DataFrame containing the permuted edgelist
    """
    if random_number_generator is None:
        random_number_generator = _get_random_number_generator()

    marker1_permute = (
        edgelist.select(["umi1", "marker_1"])
        .unique()
        .group_by("umi1")
        .first()
        .sort("umi1")
        .with_columns(
            pl.col("marker_1").shuffle(random_number_generator.integers(0, int(1e10)))  # type: ignore
        )
    )
    marker2_permute = (
        edgelist.select(["umi2", "marker_2"])
        .unique()
        .group_by("umi2")
        .first()
        .sort("umi2")
        .with_columns(
            pl.col("marker_2").shuffle(random_number_generator.integers(0, int(1e10)))  # type: ignore
        )
    )

    permuted_edgelist = edgelist.with_columns(
        marker_1=pl.col("umi1").replace_strict(
            marker1_permute["umi1"], marker1_permute["marker_1"]
        ),
        marker_2=pl.col("umi2").replace_strict(
            marker2_permute["umi2"], marker2_permute["marker_2"]
        ),
    )
    return permuted_edgelist


def permute_node_markers(
    node_markers: pd.DataFrame,
    random_number_generator: Optional[RandomNumberGenerator] = None,
    node_a_rows: Optional[pd.Series] = None,
):
    """Permute markers in a node_markers DataFrame.

    This function permutes the node_markers by shuffling the corresponding
    markers to umi column. If node_a_rows is provided, markers from COa nodes
    and COb nodes will be shuffled separately.

    :param node_markers: A DataFrame representing the node_markers (nodes as rows and markers as columns)
    :param random_number_generator: A RandomNumberGenerator instance
    :param node_a_rows: A boolean Series indicating which rows are COa nodes
    :returns: A DataFrame containing the permuted node_markers
    """
    if random_number_generator is None:
        random_number_generator = _get_random_number_generator()

    if node_a_rows is None:
        node_a_rows = pd.Series(False, index=node_markers.index)

    permuted_node_markers = node_markers.copy()

    permuted_node_markers.loc[node_a_rows, :] = (
        node_markers.loc[node_a_rows, :].sample(frac=1).values
    )
    permuted_node_markers.loc[~node_a_rows, :] = (
        node_markers.loc[~node_a_rows, :].sample(frac=1).values
    )
    return permuted_node_markers



"""BKTree implementation for fast sequence matching.

Copyright © 2024 Pixelgen Technologies AB.
"""

import numpy as np

from pixelator.pna.config import PNAAntibodyPanel


class BKTreeItem:
    """Small helper class to store the sequence and the marker ID in the BKTree."""

    __slots__ = ["id", "sequence"]

    def __init__(self, id: int, sequence: bytes):
        self.id = id
        self.sequence = sequence

    def __repr__(self):
        return f"BKTreeItem({self.id}, {self.sequence})"


from collections import deque
from operator import itemgetter

__all__ = ["hamming_distance", "BKTree"]

__version__ = "1.1"

_getitem0 = itemgetter(0)


def hamming_distance(x, y):
    """Calculate the hamming distance between two integral values.

    >>> [hamming_distance(x, 15) for x in [0, 8, 10, 12, 14, 15]]
    [4, 3, 2, 2, 1, 0]
    """
    return bin(x ^ y).count("1")


class BKTree:
    """BK-tree data structure.

    The BK-tree allows fast querying of matches that are
    "close" given a function to calculate a distance metric (e.g., Hamming
    distance or Levenshtein distance).

    Each node in the tree (including the root node) is a two-tuple of
    (item, children_dict), where children_dict is a dict whose keys are
    non-negative distances of the child to the current item and whose values
    are nodes.

    Adapter from: https://github.com/Jetsetter/pybktree
    License: MIT
    """

    def __init__(self, distance_func, items=[]):
        """Initialize a BKTree instance with given distance function.

        The distance function should be a callable that takes two items
        and returns a non-negative distance integer,

        :param distance_func: The distance function to use.
        :param items: An optional list of items to add on initialization.

        >>> tree = BKTree(hamming_distance)
        >>> list(tree)
        []
        >>> tree.distance_func is hamming_distance
        True
        >>> tree = BKTree(hamming_distance, [])
        >>> list(tree)
        []
        >>> tree = BKTree(hamming_distance, [0, 4, 5])
        >>> sorted(tree)
        [0, 4, 5]
        """
        self.distance_func = distance_func
        self.tree = None

        _add = self.add
        for item in items:
            _add(item)

    def add(self, item):
        """Add given item to this tree.

        >>> tree = BKTree(hamming_distance)
        >>> list(tree)
        []
        >>> tree.add(4)
        >>> sorted(tree)
        [4]
        >>> tree.add(15)
        >>> sorted(tree)
        [4, 15]
        """
        node = self.tree
        if node is None:
            self.tree = (item, {})
            return

        # Slight speed optimization -- avoid lookups inside the loop
        _distance_func = self.distance_func

        while True:
            parent, children = node
            distance = _distance_func(item, parent)
            node = children.get(distance)
            if node is None:
                children[distance] = (item, {})
                break

    def find(self, item, n):
        """Find items in this tree with a distance <= `n` from `item`.

         Return list of (distance, item) tuples ordered by distance.

        :param item: The item to find matches for.
        :param n: The maximum distance to consider a match.

        >>> tree = BKTree(hamming_distance)
        >>> tree.find(13, 1)
        []
        >>> tree.add(0)
        >>> tree.find(1, 1)
        [(1, 0)]
        >>> for item in [0, 4, 5, 14, 15]:
        ...     tree.add(item)
        >>> sorted(tree)
        [0, 0, 4, 5, 14, 15]
        >>> sorted(tree.find(13, 1))
        [(1, 5), (1, 15)]
        >>> sorted(tree.find(13, 2))
        [(1, 5), (1, 15), (2, 4), (2, 14)]
        >>> sorted(tree.find(0, 1000)) == [(hamming_distance(x, 0), x) for x in tree]
        True
        """
        if self.tree is None:
            return []

        candidates = deque([self.tree])
        found = []

        # Slight speed optimization -- avoid lookups inside the loop
        _candidates_popleft = candidates.popleft
        _candidates_extend = candidates.extend
        _found_append = found.append
        _distance_func = self.distance_func

        while candidates:
            candidate, children = _candidates_popleft()
            distance = _distance_func(candidate, item)
            if distance <= n:
                _found_append((distance, candidate))

            if children:
                lower = distance - n
                upper = distance + n
                _candidates_extend(
                    c for d, c in children.items() if lower <= d <= upper
                )

        found.sort(key=_getitem0)
        return found

    def __iter__(self):
        """Return iterator over all items in this tree.

        Items are yielded in arbitrary order.

        >>> tree = BKTree(hamming_distance)
        >>> list(tree)
        []
        >>> tree = BKTree(hamming_distance, [1, 2, 3, 4, 5])
        >>> sorted(tree)
        [1, 2, 3, 4, 5]
        """
        if self.tree is None:
            return

        candidates = deque([self.tree])

        # Slight speed optimization -- avoid lookups inside the loop
        _candidates_popleft = candidates.popleft
        _candidates_extend = candidates.extend

        while candidates:
            candidate, children = _candidates_popleft()
            yield candidate
            _candidates_extend(children.values())

    def __repr__(self):
        """Return a string representation of this BK-tree with a little bit of info.

        >>> BKTree(hamming_distance)
        <BKTree using hamming_distance with no top-level nodes>
        >>> BKTree(hamming_distance, [0, 4, 8, 14, 15])
        <BKTree using hamming_distance with 3 top-level nodes>
        """
        return "<{} using {} with {} top-level nodes>".format(
            self.__class__.__name__,
            self.distance_func.__name__,
            len(self.tree[1]) if self.tree is not None else "no",
        )


def hamming_distance_i8(s1: BKTreeItem, s2: BKTreeItem | bytes) -> int:
    """Calculate the byte-wise Hamming distance between two sequences."""
    b1 = np.frombuffer(s1.sequence, dtype=np.int8)
    b2 = np.frombuffer(s2.sequence if isinstance(s2, BKTreeItem) else s2, dtype=np.int8)
    return int(np.sum(b1 != b2))


def build_bktree(panel: PNAAntibodyPanel, sequence_key: str) -> BKTree:
    """Create a BKTree from the panel sequences.

    This allows us to quickly find the closest sequence to a given query sequence with up to a given distance.
    The distance function is the edit distance

    :param panel: The panel to build the tree from
    :param sequence_key: The key in the panel dataframe that contains the sequences
    :return: The BKTree
    """
    tree = BKTree(hamming_distance_i8)

    for _, row in panel.df.iterrows():
        seq = row[sequence_key].encode("ascii")
        tree.add(BKTreeItem(row.name, seq))

    return tree


def build_exact_dict_lookup(
    panel: PNAAntibodyPanel, sequence_key: str
) -> dict[bytes, str]:
    """Create a set from the panel sequences.

    This allows us to quickly find exact matches for a sequence in the panel.

    :param panel: The panel to build the lookup from
    :param sequence_key: The key in the panel dataframe that contains the sequences
    :return: The lookup table
    """
    lut = dict()

    for _, row in panel.df.iterrows():
        seq = row[sequence_key].encode("ascii")
        lut[seq] = row.name

    return lut



"""Module for collapsing a list of molecules into unique antibody links.

Copyright © 2024 Pixelgen Technologies AB.
"""

from .independent.collapser import RegionCollapser
from .paired.collapser import MoleculeCollapser

__all__ = ["MoleculeCollapser", "RegionCollapser"]



"""Functions to create an interactive qc report.

Copyright © 2022 Pixelgen Technologies AB.
"""

from .builder import PNAQCReportBuilder
from .main import create_per_sample_qc_reports
from .types import InfoAndMetrics, Metrics, QCReportData, SampleInfo

__all__ = [
    "QCReportData",
    "InfoAndMetrics",
    "Metrics",
    "SampleInfo",
    "PNAQCReportBuilder",
    "create_per_sample_qc_reports",
]



"""Functions for doing size-based cell calling.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from typing import Literal, Optional

import numpy as np
import pandas as pd
from scipy.interpolate import UnivariateSpline

from pixelator.common.annotate.constants import (
    CELL_MAX_SIZE_SMOOTHING_FACTOR,
    CELL_MIN_SIZE_SMOOTHING_FACTOR,
    DISTANCE_DEVIATION_FACTOR,
    MINIMUM_N_EDGES_CELL_SIZE,
    MINIMUM_NBR_OF_CELLS_FOR_SIZE_LIMIT,
    PRE_FILTER_LIMIT,
)

logger = logging.getLogger(__name__)


def find_component_size_limits(
    component_sizes: np.ndarray,
    direction: Literal["lower", "upper"],
) -> Optional[int]:
    """Find component size limits.

    This function will attempt to find a cutoff for a distribution of component sizes.
    The direction of the cut-off is determined by the `direction` parameter (lower for
    min size and upper for max size).

    The underlying assumption for the lower bound is that there is one distribution
    that consists of small components, that are mostly noise and another distribution
    of components that are larger, and that make up the true cell components.

    The method employed here tries to find a drop in the size of components by
    looking at the size vs rank, on a log scale, essentially trying to find
    the point where there is a sharp drop in the size.

    This is done by finding the minimum of the second derivative of the
    spline-smoothed log(size_definition) ~ log(rank). This is similar to the method
    described by Lun et al. [1]_

    The underlying assumption and method for the upper bound is the same but we
    assign each component a point in a coordinate system based on derivate2 ~ derivate1,
    and compute the distance d from origo for each component.

    We select the top 50% largest components, and find the maximum rank R
    of a component where:

      d > DOUBLET_DISTANCE_DEVIATION_FACTOR * stdev(d)

    Essentially finding components that are outliers both in the first and the
    second derivate. We then use R, to find the component with rank R, as the size
    cutoff.

    .. [1] Lun, A., Riesenfeld, S., Andrews, T. et al. EmptyDrops: distinguishing
        cells from empty droplets in droplet-based single-cell RNA sequencing

    :param component_sizes: a numpy array of component sizes
    :param direction: the direction of the cutoff, either "lower" or "upper"
    :return: the lower or upper bound cutoff
    :raises AssertionError: if the direction is not lower or upper
    :raises AssertionError: if component_sizes contain NaNs or zeros
    """

    def log_size_and_rank(df: pd.DataFrame) -> pd.DataFrame:
        """Rank components by size.

        Rank component sizes by size, sort by their rank, and compute the
        log10 of both the sizes and ranks. The input dataframe must contain
        a `size` column with the component sizes. A new dataframe is
        returned with two new columns `log10_size` and `log10_rank`.
        """
        df["rank"] = df["size"].rank(ascending=False, method="first")
        df = df.sort_values(["rank"])
        df["log10_size"] = np.log10(df["size"])
        df["log10_rank"] = np.log10(df["rank"])
        return df

    def smooth(df: pd.DataFrame, x_var: str, y_var: str) -> pd.DataFrame:
        """Create a smoothing spline.

        Calculate a smoothing spline of df[x_var] ~ df[y_var]
        to make it possible to calculate a less unstable derivate.
        The input dataframe must contain the `x_var` and `y_var`
        columns, a new column `smooth` is added to the returned
        dataframe.
        """
        spline_func = UnivariateSpline(
            x=df[x_var],
            y=df[y_var],
            s=spline_smoothing_factor,
            check_finite=True,
        )
        df["smooth"] = spline_func(df[x_var])
        return df

    def derivatives(df: pd.DataFrame, x_var: str) -> pd.DataFrame:
        """Calculate first and second derivatives.

        Calculate the first and second derivatives of the smoothed
        `x_var` variable. The input dataframe must contain the
        `x_var` and `smooth` columns. The returned dataframe
        contains two new columns `der1` and `der2` with the
        computed derivatives.
        """
        df["der1"] = df["smooth"].diff() / df[x_var].diff()
        df["der2"] = df["der1"].diff() / df[x_var].diff()
        return df

    def find_der1_vs_der2_outliers(df: pd.DataFrame) -> pd.Series:
        """Find the distance from origo to each component.

        Find the distance from origo to each component in
        the space df[der1] ~ df[der2], then try to find
        outliers in the upper part of component ranks, by
        looking at the standard deviation of the distances.
        Finding the components that maximizes the
        rank and is an outlier. The input dataframe must
        contain the `der1` and `der2` columns and the returned
        dataframe will contain the `distance` and `rank` columns.
        The functions return a boolean Series where every outlier
        evaluates to True.
        """
        df["distance"] = np.linalg.norm(df[["der1", "der2"]].to_numpy(), axis=1)
        stddev_distance = np.std(df["distance"])
        largest_half = df["rank"] < (np.max(df["rank"]) / 2)
        extrem_distance = df["distance"] > DISTANCE_DEVIATION_FACTOR * stddev_distance
        rank = np.max(df[extrem_distance & largest_half]["rank"])
        return df["rank"] == rank - 1

    def minimum_der2(df: pd.DataFrame) -> pd.Series:
        """Find argmin element.

        The function returns a boolean Series where the global minimum of `der2` evaluates
        to True. The input dataframe must contain the `der2` column.
        """
        return df["der2"] == np.nanmin(df["der2"])

    if direction not in ["lower", "upper"]:
        raise AssertionError(f"direction must be lower or upper, got: {direction}")

    if np.isnan(component_sizes).any():
        raise AssertionError("the component sizes contains NaN values")

    if (component_sizes <= 0).any():
        raise AssertionError("the component sizes contains zeros or negative values")

    n_unique = len(np.unique(component_sizes))
    logger.debug(
        "Attempting to find size %s limit on %i component sizes where %i are unique",
        direction,
        len(component_sizes),
        n_unique,
    )

    # add a pre-filter step to make the method more robust
    pre_filter = component_sizes > PRE_FILTER_LIMIT
    logger.debug(
        (
            "Pre-filtering %s (%s%%) component sizes with size smaller than %s "
            "before trying to find a %s size limit"
        ),
        np.sum(pre_filter),
        round((np.sum(pre_filter) / len(pre_filter)) * 100, 0),
        PRE_FILTER_LIMIT,
        direction,
    )
    component_sizes = component_sizes[pre_filter]

    # check for the sizes that are unique
    if n_unique < MINIMUM_NBR_OF_CELLS_FOR_SIZE_LIMIT:
        logger.warning(
            "Too few unique component sizes (%i) to find a %s size limit",
            len(component_sizes),
            direction,
        )
        return None

    if direction == "lower":
        x_var = "log10_rank"
        y_var = "log10_size"
        determining_func = minimum_der2
        spline_smoothing_factor = CELL_MIN_SIZE_SMOOTHING_FACTOR
    else:
        x_var = "rank"
        y_var = "log10_size"
        spline_smoothing_factor = CELL_MAX_SIZE_SMOOTHING_FACTOR
        determining_func = find_der1_vs_der2_outliers

    # compute the smoothed derivatives of the log-rank sizes
    df = pd.DataFrame({"size": component_sizes})
    df = log_size_and_rank(df)
    try:
        df = smooth(df, x_var, y_var)
    except ValueError as err:
        logger.warning(
            "Spline smoothing failed with error %s, no size limit will be returned",
            str(err),
        )
        return None
    df = derivatives(df, x_var)

    # derive the cutoff
    df["cutoff"] = determining_func(df)
    potential_bounds = df[df["cutoff"]]["size"].tolist()
    bound = potential_bounds[0] if potential_bounds else None

    logger.debug("Size limit of %i found", bound)

    if direction == "lower" and bound is not None:
        if bound < MINIMUM_N_EDGES_CELL_SIZE:
            logger.warning(
                "Dynamic minimum component size found is below the minimum size threshold of %s edges. "
                "Using that as the minimum size to avoid downstream problems.",
                MINIMUM_N_EDGES_CELL_SIZE,
            )
            bound = MINIMUM_N_EDGES_CELL_SIZE

    return bound



"""Routines for printing a report.

Copyright © 2024 Pixelgen Technologies AB.
"""

import typing
from typing import Any, Dict, Optional

import pydantic
from cutadapt.report import Statistics

from pixelator.pna.read_processing.statistics import HasCustomStatistics
from pixelator.pna.report.models.base import SampleReport


class Q30Statistics(pydantic.BaseModel):
    """The result of the quality statistics."""

    total: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the complete amplicon.",
    )
    umi1: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the UMI1 region.",
    )
    pid1: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the PID1 region.",
    )
    lbs1: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the LBS1 region.",
    )
    uei: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the UEI region.",
    )
    lbs2: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the LBS2 region.",
    )
    pid2: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the PID2 region.",
    )
    umi2: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the UMI2 region.",
    )


class BasesCountStatistics(pydantic.BaseModel):
    """The result of the base count statistics."""

    input: int
    input_read1: int = pydantic.Field(
        ..., description="The number of input bases in read1."
    )
    input_read2: int = pydantic.Field(
        ..., description="The number of input bases in read2."
    )
    quality_trimmed: int = pydantic.Field(
        ..., description="The total number of quality trimmed bases."
    )
    quality_trimmed_read1: int = pydantic.Field(
        ..., description="The number of quality trimmed bases in read1."
    )
    quality_trimmed_read2: Optional[int] = pydantic.Field(
        ..., description="The number of quality trimmed bases in read2."
    )
    output: int = pydantic.Field(..., description="The total number of output bases.")


class AmpliconSampleReport(SampleReport):
    """Model for an amplicon sample report."""

    report_type: typing.Literal["amplicon"] = "amplicon"

    input_reads: int = pydantic.Field(
        ..., description="The total number of input reads in the amplicon stage."
    )

    output_reads: int = pydantic.Field(
        ...,
        description="The number of reads that passed the filters in the amplicon stage.",
    )

    passed_missing_uei_reads: int = pydantic.Field(
        ...,
        description="The number of reads that passed but do not have a recognisable UEI region.",
    )

    passed_partial_uei_reads: int = pydantic.Field(
        ...,
        description="The number of reads that passed but do not have a fully sequenced UEI region.",
    )

    passed_missing_lbs1_anchor: int = pydantic.Field(
        ...,
        description="The number of reads that passed but do not have a recognisable LBS1 anchor.",
    )

    failed_too_many_n_reads: int = pydantic.Field(
        ...,
        description="The number of reads discarded because they contain too many ambiguous bases.",
    )

    @pydantic.computed_field(
        return_type=int,
        description=(
            "The number of reads discarded because they are missing or have a partial region."
            "This is the sum of the failed or partial upi1/umi1 and upi2/umi2 reads."
        ),
    )
    def failed_invalid_amplicon_reads(self) -> int:
        """Calculate the number of reads discarded because they are missing or have a partial region."""
        return (
            self.failed_partial_upi1_umi1_reads
            + self.failed_partial_upi2_umi2_reads
            + self.failed_missing_upi1_umi1_reads
            + self.failed_missing_upi2_umi2_reads
        )

    failed_partial_upi1_umi1_reads: int = pydantic.Field(
        ...,
        description="The number of reads discarded because they have a partial UPI1/UMI1 region.",
    )

    failed_partial_upi2_umi2_reads: int = pydantic.Field(
        ...,
        description="The number of reads discarded because they have a partial UPI2/UMI2 region.",
    )

    failed_missing_upi1_umi1_reads: int = pydantic.Field(
        ...,
        description="The number of reads discarded because they are missing UPI1/UMI1 sequences.",
    )

    failed_missing_upi2_umi2_reads: int = pydantic.Field(
        ...,
        description="The number of reads discarded because they are missing UPI2/UMI2 sequences.",
    )

    total_failed_reads: int = pydantic.Field(
        ...,
        description="The total number of reads that failed the filters and are discarded.",
    )

    q30_statistics: Q30Statistics = pydantic.Field(
        ..., description="Q30 statistics for the amplicon stage."
    )

    basepair_counts: BasesCountStatistics = pydantic.Field(
        ..., description="Base count statistics for the amplicon stage."
    )

    @pydantic.computed_field(  # type: ignore
        description="The fraction of reads that was discarded by the amplicon step.",
        return_type=float,
    )
    @property
    def fraction_discarded_reads(self) -> float:
        """Calculate the fraction of reads that were discarded by the amplicon step."""
        if self.input_reads == 0:
            return 0
        return self.total_failed_reads / self.input_reads


class AmpliconStatistics(Statistics):
    """Statistics for the amplicon stage."""

    FILTERS = {
        "too_many_n": "with too many N",
        "invalid_amplicon": "with missing amplicon regions",
    }

    def __init__(self) -> None:
        """Initialize the AmpliconStatistics object."""
        super().__init__()
        self.paired: bool | None = None
        self._custom_stats: dict[str, Any] = dict()

    def __iadd__(self, other: Any):
        """Merge statistics from another object into this one."""
        if other.paired is None:
            other.paired = self.paired
        super().__iadd__(other)
        if hasattr(other, "_custom_stats"):
            for name, value in other._custom_stats.items():
                if name in self._custom_stats:
                    self._custom_stats[name] += value
                else:
                    self._custom_stats[name] = value

        return self

    def as_json(self, gc_content: float = 0.5, one_line: bool = False) -> Dict:
        """Return a dict representation suitable for dumping in JSON format."""
        return self.as_dict()

    def as_dict(self) -> Dict:
        """Return a dict representation of the class."""
        filtered = {name: self.filtered.get(name) for name in self.FILTERS.keys()}
        filtered_total = sum(self.filtered.values())
        written_reads = self.read_length_statistics.written_reads()
        written_bp = self.read_length_statistics.written_bp()
        assert (written_reads + filtered_total) == self.n

        q30_regions_report = None
        amplicon_report = None
        if "quality_profile" in self._custom_stats:
            q30_regions_report = self._custom_stats["quality_profile"].collect()

        if "amplicon" in self._custom_stats:
            amplicon_report = self._custom_stats["amplicon"].collect()

        # type: ignore
        q30_stats = Q30Statistics(
            total=q30_regions_report["fraction_q30_total"],  # type: ignore
            umi1=q30_regions_report["fraction_q30_umi1"],  # type: ignore
            pid1=q30_regions_report["fraction_q30_pid1"],  # type: ignore
            lbs1=q30_regions_report["fraction_q30_lbs1"],  # type: ignore
            uei=q30_regions_report["fraction_q30_uei"],  # type: ignore
            lbs2=q30_regions_report["fraction_q30_lbs2"],  # type: ignore
            pid2=q30_regions_report["fraction_q30_pid2"],  # type: ignore
            umi2=q30_regions_report["fraction_q30_umi2"],  # type: ignore
        )
        basepair_counts = BasesCountStatistics(
            input=self.total,  # type: ignore
            input_read1=self.total_bp[0],  # type: ignore
            input_read2=self.total_bp[1],  # type: ignore
            quality_trimmed=self.quality_trimmed,  # type: ignore
            quality_trimmed_read1=self.quality_trimmed_bp[0],  # type: ignore
            quality_trimmed_read2=self.quality_trimmed_bp[1],  # type: ignore
            output=self.total_written_bp,  # type: ignore
        )

        # type: ignore
        report = AmpliconSampleReport(
            sample_id="",
            report_type="amplicon",
            product_id="single-cell-pna",
            input_reads=self.n,
            output_reads=written_reads,
            total_failed_reads=filtered_total,
            passed_missing_uei_reads=amplicon_report["passed_missing_eui_reads"],  # type: ignore
            passed_partial_uei_reads=amplicon_report["passed_partial_uei_reads"],  # type: ignore
            passed_missing_lbs1_anchor=amplicon_report["passed_missing_lbs1_anchor"],  # type: ignore
            failed_too_many_n_reads=self.filtered["too_many_n"],  # type: ignore
            failed_partial_upi1_umi1_reads=amplicon_report[
                "failed_partial_upi1_umi1_reads"
            ],  # type: ignore
            failed_partial_upi2_umi2_reads=amplicon_report[
                "failed_partial_upi2_umi2_reads"
            ],  # type: ignore
            failed_missing_upi1_umi1_reads=amplicon_report[
                "failed_missing_upi1_umi1_reads"
            ],  # type: ignore
            failed_missing_upi2_umi2_reads=amplicon_report[
                "failed_missing_upi2_umi2_reads"
            ],  # type: ignore
            q30_statistics=q30_stats,
            basepair_counts=basepair_counts,
        )
        return report.model_dump(
            mode="python", exclude={"sample_id", "report_type", "product_id"}
        )

    def _collect_step(self, step):
        super()._collect_step(step)
        if isinstance(step, HasCustomStatistics):
            name = step.get_statistics_name()
            if name in self._custom_stats:
                self._custom_stats[name] += step.get_statistics()
            else:
                self._custom_stats[name] = step.get_statistics()

    def collect(
        self,
        n: int,
        total_bp1: int,
        total_bp2: int | None,
        modifiers,
        steps,
        set_paired_to_none: bool = False,
    ):
        """Enable stats.paired to be set to None when unknown."""
        stats = super().collect(n, total_bp1, total_bp2, modifiers, steps)
        if set_paired_to_none:
            stats.paired = None
        return stats



"""Console script for pixelator (annotate).

Copyright © 2022 Pixelgen Technologies AB.
"""

import click

from pixelator.common.annotate.constants import MINIMUM_N_EDGES_CELL_SIZE
from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.annotate import annotate_components
from pixelator.mpx.cli.common import logger, output_option
from pixelator.mpx.config import config, load_antibody_panel


@click.command(
    "annotate",
    short_help="filter, annotate and call cells from an edge list",
    options_metavar="<options>",
)
@click.argument(
    "parquet_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="parquet",
)
@click.option(
    "--panel",
    required=True,
    type=str,
    help=(
        "A key of a panel file in the config, "
        "or a csv file with the antibody panel conjugations"
    ),
)
@click.option(
    "--min-size",
    default=None,
    required=False,
    type=click.INT,
    show_default=False,
    help="The minimum size (edges) a component must have (default is disabled). Note that this cannot be set at the same time as --dynamic-filter.",
)
@click.option(
    "--max-size",
    default=None,
    required=False,
    type=click.INT,
    show_default=False,
    help="The maximum size (edges) a component must have (default is disabled). Note that this cannot be set at the same time as --dynamic-filter.",
)
@click.option(
    "--dynamic-filter",
    required=False,
    default=None,
    type=click.Choice(["both", "min", "max"]),
    help=(
        "Enable the dynamic component size filters. The following modes are available: "
        "both/max/min. both: estimates both minimum and maximum component size, min: estimates the minimum component "
        f"size (or uses {MINIMUM_N_EDGES_CELL_SIZE} edges, whichever is smallest), "
        "max: estimates the maximum component size. Note that this cannot be set at the same time as --min-size or --max-size."
    ),
)
@click.option(
    "--aggregate-calling",
    default=False,
    is_flag=True,
    help=(
        "Enable aggregate calling, information on "
        "potential aggregates will be added to the output data"
    ),
)
@output_option
@click.pass_context
@timer
def annotate(
    ctx,
    parquet_file,
    panel,
    min_size,
    max_size,
    dynamic_filter,
    aggregate_calling,
    output,
):
    """
    Filter, annotate and call cells from an edge list
    """
    input_files = [parquet_file]
    # log input parameters
    log_step_start(
        "annotate",
        input_files=input_files,
        panel=panel,
        output=output,
        min_size=min_size,
        max_size=max_size,
        dynamic_filter=dynamic_filter,
        aggregate_calling=aggregate_calling,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files, allowed_extensions="parquet")

    # sanity check on thresholds and input parameters
    if min_size is not None and min_size < 0:
        raise click.ClickException("--min-size cannot be less than 0")
    if max_size is not None and max_size < 0:
        raise click.ClickException("--max-size cannot be less than 0")
    if max_size is not None and min_size is not None and max_size < min_size:
        raise click.ClickException("--max-size cannot be less than --min-size")

    if min_size is not None or max_size is not None:
        if dynamic_filter is not None:
            raise click.ClickException(
                "Cannot set --dynamic-filter and --min-size or --max-size at the same time"
            )

    # warn if both --dynamic-filter and hard-coded sizes are input
    if min_size is not None and dynamic_filter in ["min", "both"]:
        logger.warning(
            "--dynamic-filter will overrule the value introduced in --min-size"
        )

    if max_size is not None and dynamic_filter in ["max", "both"]:
        logger.warning(
            "--dynamic-filter will overrule the value introduced in --max-size"
        )

    # create output folder if it does not exist
    annotate_output = create_output_stage_dir(output, "annotate")

    # load marker panel
    panel = load_antibody_panel(config, panel)

    # compute graph/components using parallel processing
    logger.info(f"Computing annotation for file {parquet_file}")

    clean_name = get_sample_name(parquet_file)
    metrics_file = annotate_output / f"{clean_name}.report.json"

    write_parameters_file(
        ctx,
        annotate_output / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-mpx annotate",
    )

    annotate_components(
        input=str(parquet_file),
        panel=panel,
        output=str(annotate_output),
        output_prefix=clean_name,
        metrics_file=str(metrics_file),
        min_size=min_size,
        max_size=max_size,
        dynamic_filter=dynamic_filter,
        aggregate_calling=aggregate_calling,
        verbose=ctx.obj["VERBOSE"],
    )



"""Copyright © 2024 Pixelgen Technologies AB."""

import logging

import duckdb
import numpy as np
import pandas as pd
import polars as pl
from anndata import AnnData

from pixelator.pna.config import PNAAntibodyPanel

logger = logging.getLogger(__name__)


def calculate_antibody_metrics(counts_df):
    """Calculate antibody metrics from a count matrix, can be used to set/update adata.vars."""
    total_antibody = pd.Series(counts_df.sum(axis=0), name="antibody_count")
    relative_antibody = pd.Series(
        total_antibody / total_antibody.sum(), name="antibody_pct"
    )
    components_detected = pd.Series((counts_df > 0).sum(axis=0), name="components")

    return pd.concat([total_antibody, relative_antibody, components_detected], axis=1)


def add_panel_information(adata, panel):
    """Add panel data to var."""
    adata.var = adata.var.join(panel.df, how="left")

    adata.uns["panel_metadata"] = {
        "name": panel.name,
        "aliases": panel.aliases,
        "description": panel.description,
        "version": panel.version,
        "panel_columns": list(panel.df.columns),
    }

    return adata


def pna_edgelist_to_anndata(
    pixel_connection: duckdb.DuckDBPyConnection, panel: PNAAntibodyPanel
) -> AnnData:
    """Build an AnnData object from a DuckDB connection to a pixel file and a panel object.

    Parameters
    ----------
    pixel_connection : duckdb.DuckDBPyConnection
        A DuckDB connection to a pixel file. The connection must contain an 'edgelist' table
        with the required columns (e.g., component, marker_1, marker_2, umi1, umi2, read_count).
    panel : PNAAntibodyPanel
        The antibody panel object containing marker metadata.

    Returns
    -------
    AnnData
        An AnnData object with counts and panel information.

    Notes
    -----
    Assumes that the 'edgelist' table exists in the DuckDB connection and contains the necessary columns.

    """
    logger.debug("Constructing counts matrix.")

    marker_names = [f"'{m}'" for m in panel.markers]
    marker_names_sql = ", ".join(marker_names)
    node_counts_df = (
        pixel_connection.execute(f"""
        SELECT *
        FROM (
            WITH counts_df_long AS (
                WITH
                    marker_1_counts AS (
                        SELECT component, marker_1 AS marker, COUNT(DISTINCT umi1) AS marker_1_count
                        FROM edgelist
                        GROUP BY component, marker_1),
                    marker_2_counts AS (
                        SELECT component, marker_2 AS marker, COUNT(DISTINCT umi2) AS marker_2_count
                        FROM edgelist
                        GROUP BY component, marker_2
                    )
                SELECT
                    COALESCE(a.component, b.component) AS component,
                    COALESCE(a.marker, b.marker) AS marker,
                    COALESCE(a.marker_1_count, 0) AS marker_1_count,
                    COALESCE(b.marker_2_count, 0) AS marker_2_count,
                    COALESCE(a.marker_1_count, 0) + COALESCE(b.marker_2_count, 0) AS count
                FROM marker_1_counts a
                FULL OUTER JOIN marker_2_counts b
                    ON a.component = b.component AND a.marker = b.marker
            )
            PIVOT counts_df_long
            ON marker IN ({marker_names_sql})
            USING SUM(count)
            GROUP BY component
        )
    """)
        .df()
        .fillna(0)
    )

    node_counts_df.set_index("component", inplace=True)
    node_counts_df = node_counts_df.reindex(columns=panel.markers, fill_value=0)
    node_counts_df = node_counts_df.astype("uint32")

    # compute components metrics (obs) and re-index
    logger.debug("Computing component metrics.")

    components_metrics_df = pixel_connection.execute(f"""
            WITH
                marker_1_counts AS (
                    SELECT component, marker_1 AS marker, COUNT(DISTINCT umi1) AS marker_1_count
                    FROM edgelist
                    GROUP BY component, marker_1),
                marker_2_counts AS (
                    SELECT component, marker_2 AS marker, COUNT(DISTINCT umi2) AS marker_2_count
                    FROM edgelist
                    GROUP BY component, marker_2
                ),
                component_marker_counts AS (
                    SELECT
                        COALESCE(a.component, b.component) AS component,
                        COALESCE(a.marker_1_count, 0) AS marker_1_count,
                        COALESCE(b.marker_2_count, 0) AS marker_2_count
                    FROM marker_1_counts a
                    FULL OUTER JOIN marker_2_counts b
                        ON a.component = b.component AND a.marker = b.marker
                    ),
                component_umi AS (
                    SELECT
                        component,
                        SUM(marker_1_count) AS n_umi1,
                        SUM(marker_2_count) AS n_umi2
                    FROM component_marker_counts
                    GROUP BY component
                ),
                edge_counts AS (
                    SELECT component, COUNT(*) AS n_edges, SUM(read_count) AS reads_in_component
                    FROM edgelist
                    GROUP BY component
                )
            SELECT
                u.component,
                n_umi1,
                n_umi2,
                e.n_edges,
                e.reads_in_component,
                (n_umi1 + n_umi2) AS n_umi
            FROM component_umi u
            LEFT JOIN edge_counts e ON u.component = e.component
            ORDER BY u.component
    """).df()

    n_antibodies = pd.Series(
        (node_counts_df != 0).sum(axis=1),
        index=node_counts_df.index,
        name="n_antibodies",
        dtype=np.uint32,
    )
    components_metrics_df.set_index("component", inplace=True)
    components_metrics_df = components_metrics_df.join(n_antibodies)

    components_metrics_df = components_metrics_df.astype(
        {
            "n_umi": np.uint64,
            "n_umi1": np.uint64,
            "n_umi2": np.uint64,
            "n_edges": np.uint64,
            "n_antibodies": np.uint32,
            "reads_in_component": np.uint64,
        }
    )
    components_metrics_df = components_metrics_df.reindex(index=node_counts_df.index)
    components_metrics_df.index = components_metrics_df.index.astype(str)

    # compute antibody metrics (var) and re-index
    logger.debug("Computing antibody metrics.")
    antibody_metrics_df = calculate_antibody_metrics(counts_df=node_counts_df)
    antibody_metrics_df = antibody_metrics_df.reindex(index=panel.markers, fill_value=0)
    # Do a dtype conversion of the columns here since AnnData cannot handle
    # a pyarrow arrays.
    antibody_metrics_df = antibody_metrics_df.astype(
        {"antibody_count": "int64", "antibody_pct": "float32"}
    )
    antibody_metrics_df.index = antibody_metrics_df.index.astype(str)

    # create AnnData object
    node_counts_df.index = node_counts_df.index.astype(
        str
    )  # anndata requires indexes to be strings
    logger.debug("Building AnnData instance.")
    adata = AnnData(
        X=node_counts_df,
        obs=components_metrics_df,
        var=antibody_metrics_df,
    )

    adata = add_panel_information(adata, panel)

    # find fraction of isotype markers in cell
    total_marker_counts = node_counts_df.sum(axis=1)
    isotype_markers = adata.var[adata.var["control"]].index
    isotype_counts = node_counts_df[isotype_markers].sum(axis=1)
    adata.obs["isotype_fraction"] = isotype_counts / total_marker_counts

    # This is set to preserve backwards compatibility with downstream reports that may expect it.
    # Eventually we should be able to remove this.
    adata.obs["intracellular_fraction"] = 0.0

    return adata
